@@ -27,7 +27,7 @@ training sentiment analysis tasks on the IMDB dataset::
     data.FilterByLength(max_length=2048, length_keys=[0]),
     data.BucketByLength(boundaries=[  32, 128, 512, 2048],
                         batch_sizes=[128,  32,   8,    2, 1],
-                        length_keys=[0])
+                        length_keys=[0]),
     data.AddLossWeights()
   )
 
@@ -47,7 +47,7 @@ The other combinators, like Tokenize and Shuffle, take a generator and output
 another generator, in this way converting tuples into other tuples or mixing
 the training stream. For example, Tokenize(..., keys=[0]) will tokenize the
 first element of the tuple - and in this way convert it from text to a tensor of
-integers. Shuffle will not change the exmples, but will randomize their order.
+integers. Shuffle will not change the examples, but will randomize their order.
 
 Note that all elements in the data pipeline are just functions on generators,
 so you can use python's `map` and `filter` and other native functions too.
@@ -82,11 +82,12 @@ from trax.fastmath import numpy as jnp
 
 
 def Serial(*fns):  # pylint: disable=invalid-name
-  """Creates an input pipeline by running all functions one after another."""
-  generator = None
-  for f in fastmath.tree_flatten(fns):
-    generator = f(generator)
-  return generator
+  """Combines generator functions into one that runs them in turn."""
+  def composed_fns(generator=None):
+    for f in fastmath.tree_flatten(fns):
+      generator = f(generator)
+    return generator
+  return composed_fns
 
 
 def Log(n_steps_per_example=1, only_shapes=True):  # pylint: disable=invalid-name
@@ -152,9 +153,9 @@ def Shuffle(queue_size=1024):  # pylint: disable=invalid-name
 
 def batch(generator, batch_size):
   """Batch and pad generator as in tf.data.Dataset.padded_batch."""
+  if batch_size <= 0:
+    raise ValueError(f'Batch size must be positive, but is {batch_size}.')
   buf = []
-  # TODO(lukaszkaiser): convert to ValueError
-  assert batch_size > 0, f'Batch size must be positive, but is {batch_size}'
   for example in generator:
     buf.append(example)  # Examples are tuples of tensors.
     if len(buf) == batch_size:
@@ -178,7 +179,7 @@ def pad_to_max_dims(tensors, boundary=None, strict_pad_on_len=False):
   to (3, 10) both and the returned tensor will have shape (2, 3, 10).
 
   When boundary is specified, we try to pad all unknown dimensions to boundary
-  if possible, which can help reduce the number of different shapes occuring
+  if possible, which can help reduce the number of different shapes occurring
   in the tensors and speed up XLA compilation. So, for example, a pair of
   tensors of shapes (8, 10), (8, 9) with boundary=12 will be padded to (8, 12).
 
@@ -292,7 +293,7 @@ def _buckets_for_length(bucket_length, batch_size, max_eval_length, n_devices,
                         batch_size // 16, 1]
   if not training:
     # The last bucket batch size is always 1, but the one-but-last is
-    # sized to accomodate the final length = bucket_boundaries[-1], which
+    # sized to accommodate the final length = bucket_boundaries[-1], which
     # we changed for eval above -- so adjusting here too.
 
     # Resize if needed, since bucket_batch_sizes may not be the same size
@@ -381,7 +382,7 @@ def add_loss_weights(generator, id_to_mask=None):
   - If the stream consists of pairs `(inputs, targets)`, a loss mask is added
     that is creates as a tensor of ones of the same shape as targets.
   - If `id_to_mask` is not `None`, and the stream (after the previous point)
-    has triples `(inputs, targets, weights)`, the weights are multipled by a
+    has triples `(inputs, targets, weights)`, the weights are multiplied by a
     0/1 mask that is 0 iff targets is equal to `id_to_mask` (1 otherwise).
 
   Args:
@@ -698,6 +699,43 @@ def sequence_copy_inputs(
   )
 
 
+@gin.configurable()
+def simple_sequence_copy_inputs(
+    vocab_size=gin.REQUIRED, batch_size=gin.REQUIRED, train_length=gin.REQUIRED,
+    eval_min_length=gin.REQUIRED, eval_max_length=gin.REQUIRED,
+    pad_to_multiple=32):
+  """Inputs for the sequence copy problem: w for w in [1..vocab_size-1]*.
+
+  Args:
+    vocab_size: how many symbols to use.
+    batch_size: how large are the batches.
+    train_length: maximum length of w for training.
+    eval_min_length: minimum length of w for eval.
+    eval_max_length : maximum length of w for eval.
+    pad_to_multiple: int, pad length to be multiple of this number.
+
+  Returns:
+    trax.inputs.Inputs
+  """
+  def random_minibatches(length_list):
+    """Generate a stream of random mini-batches."""
+    while True:
+      length = random.choice(length_list)
+      x = np.random.randint(low=1, high=vocab_size-1,
+                            size=(batch_size, length))
+      loss_weights = np.ones((batch_size, length))
+      x = _pad_to_multiple_of(x, pad_to_multiple, 1)
+      loss_weights = _pad_to_multiple_of(loss_weights, pad_to_multiple, 1)
+      yield (x, x, loss_weights)  # Here inputs and targets are the same.
+
+  train_lengths = list(range(1, train_length + 1))
+  eval_lengths = list(range(eval_min_length, eval_max_length + 1))
+  return Inputs(
+      train_stream=lambda _: random_minibatches(train_lengths),
+      eval_stream=lambda _: random_minibatches(eval_lengths)
+  )
+
+
 def lower_endian_to_number(l, base):
   """Helper function: convert a list of digits in the given base to a number."""
   return sum([d * (base**i) for i, d in enumerate(l)])
