@@ -68,13 +68,14 @@ def _t5_gin_config():
   max_input_length = 50
 
   # What preprocessors to apply - we select a random chunk of the document if
-  # it exceeds a certain lengths (`select_random_chunk`), the concat multiple
-  # documents together to reduce padding (`reduce_concat_tokens`), then split
-  # up long examples (`split_tokens`) and finally the denoising objective
-  # (`denoise`).
+  # it exceeds a certain lengths (`select_random_chunk`), then split up long
+  # examples (`split_tokens`) and finally the denoising objective (`denoise`).
+  #
+  # In addition to this T5 concates multiple documents together to reduce
+  # padding (`reduce_concat_tokens`) after `select_random_chunk`, but we skip
+  # that since we don't do sequence packing.
   gin.bind_parameter('unsupervised.preprocessors', [
       t5_processors.select_random_chunk,
-      t5_processors.reduce_concat_tokens,
       t5_processors.split_tokens,
       t5_processors.denoise,
   ])
@@ -324,7 +325,7 @@ class TFInputsTest(tf.test.TestCase):
         bare_preprocess_fn=tf_inputs.c4_bare_preprocess_fn)
 
   def test_generic_text_dataset_preprocess_fn(self):
-    dataset = _load_dataset('squad')
+    dataset = _load_dataset('squad/v1.1:2.0.0')
 
     example, = tfds.as_numpy(dataset.take(1))
 
