@@ -107,17 +107,49 @@ def _tf_pmap(*args, **kwargs):
   return tf_np_extensions.pmap(*args, **kwargs)
 
 
+def _tf_grad(f, **kwargs):
+  """Grad with support for argnums."""
+  argnums = kwargs.pop('argnums', 0)
+  if argnums != 0:
+    def g(*args, **kwargs):
+      args = list(args)
+      args[0], args[argnums] = args[argnums], args[0]
+      return f(*args, **kwargs)
+  else:
+    g = f
+  grad_g = tf_np_extensions.grad(g, **kwargs)
+  if argnums == 0:
+    return grad_g
+  def grad_f(*args, **kwargs):
+    args = list(args)
+    args[0], args[argnums] = args[argnums], args[0]
+    return grad_g(*args, **kwargs)
+  return grad_f
+
+
+def _fold_in(rng, d):
+  """Equivalent of jax.random.fold_in."""
+  # TODO(lukaszkaiser): verify that this function has good randomness
+  # properties or switch to an implementation equivalent to JAX.
+  _, rng = tf_np_extensions.split(rng + tf_np.sum(d).astype(tf_np.int64), 2)
+  return rng
+
+
 TF_BACKEND = {
-    'name': 'tf',
+    'name': 'tensorflow-numpy',
     'np': tf_np,
     'jit': _tf_jit,
     'stop_gradient': tf_np_extensions.stop_gradient,
-    'grad': tf_np_extensions.grad,
+    'grad': _tf_grad,
     'vjp': tf_np_extensions.vjp,
     'custom_grad': tf_np_extensions.custom_grad,
     'abstract_eval': tf_abstract_eval,
     'expit': tf_np_extensions.expit,
     'erf': tf_np_extensions.erf,
+    'index_update': tf_np_extensions.index_update,
+    'index_add': tf_np_extensions.index_add,
+    'index_min': tf_np_extensions.index_min,
+    'index_max': tf_np_extensions.index_max,
     'logsumexp': tf_np_extensions.logsumexp,
     'conv': tf_np_extensions.conv,
     'lt': lambda x, y: x < y,
@@ -130,11 +162,14 @@ TF_BACKEND = {
     'random_bernoulli': tf_np_extensions.bernoulli,
     'random_get_prng': tf_np_extensions.prng,
     'random_split': tf_np_extensions.split,
+    'random_fold_in': _fold_in,
     # TODO(wangpeng): See whether and how to support `remat`
     'remat': lambda f: f,
     'scan': tf_np_extensions.scan,
-    'dataset_as_numpy': tf_np_extensions.dataset_as_numpy,
+    # TODO(wangpeng): can we make extensions ds_as_numpy compatible with data?
+    # 'dataset_as_numpy': tf_np_extensions.dataset_as_numpy,
     'device_count': lambda: max(len(tf_np_extensions.accelerators()), 1),
     'pmap': _tf_pmap,
     'psum': tf_np_extensions.psum,
+    'vmap': tf_np_extensions.vmap,
 }
