@@ -439,7 +439,9 @@ class Layer:
 
   def weights_and_state_signature(self, input_signature):
     """Return a pair containing the signatures of weights and state."""
+    rng = self.rng
     abstract_init = fastmath.abstract_eval(self.init)
+    self.rng = rng
     return abstract_init(input_signature)
 
   @property
@@ -507,7 +509,7 @@ class Layer:
         outputs = self.forward(x)
         s = self.state
       else:
-        outputs, s = self._do_custom_gradients(x, weights, state, rng=rng)
+        outputs, s = self._do_custom_gradients(x)
         self.state = s
       self._rng = old_rng
       if not use_cache:
@@ -559,35 +561,37 @@ class Layer:
                        trace) from None
 
   # pylint: disable=protected-access
-  def _do_custom_gradients(self, x, weights, state, rng):
+  def _do_custom_gradients(self, x):
     """Calls this layer for a forward pass, but with custom gradients."""
 
-    def _do_forward(y, weights):
+    def _f(state, rng, y, weights):
       old_weights, old_state, old_rng = self.weights, self.state, self._rng
-      self.weights = weights
+      self.weights, self.state, self._rng = weights, state, rng
       res = self.forward(y)
       s = self.state
       self.weights, self.state, self._rng = old_weights, old_state, old_rng
       return res, s
 
-    def do_forward_vjp(y, weights):
-      """Custom gradient (vjp) function."""
+    def _f_fwd(state, rng, y, weights):
       old_weights, old_state, old_rng = self.weights, self.state, self._rng
-      self.weights = weights
-      output = self.forward(y)
-      new_state = self.state
+      self.weights, self.state, self._rng = weights, state, rng
+      res = self.forward(y)
+      s = self.state
       self.weights, self.state, self._rng = old_weights, old_state, old_rng
-      def vjpfun(grad):
-        grad = grad[0]  # Ignore dummy gradient wrt state.
-        res = self.backward(y, output, grad, weights, state, new_state, rng)
-        return res
-      return (output, new_state), vjpfun
+      return (res, s), (state, rng, y, res, weights, s)
 
-    do_forward = fastmath.custom_grad(do_forward_vjp, _do_forward)
+    def _f_bwd(residual, grad):
+      """Custom gradient function."""
+      state, rng, y, output, weights, new_state = residual
+      grad = grad[0]  # Ignore dummy gradient wrt state.
+      out = self.backward(y, output, grad, weights, state, new_state, rng)
+      return (None, None, *out)
 
-    output, state = do_forward(x, weights)
+    do_forward = fastmath.custom_vjp(_f, _f_fwd, _f_bwd, nondiff_argnums=(0, 1))
+
+    output, state = do_forward(self.state, self._rng, x, self.weights)
     # TODO(lukaszkaiser): Investigate why we need this stop_gradient
-    state = fastmath.stop_gradient(state)
+    # state = fastmath.stop_gradient(state)
     return output, state
 
 
@@ -714,7 +718,7 @@ class LayerError(Exception):
     self._caller = caller  # Python inspect object with init caller info.
     self._traceback = traceback_string
     self._input_signature = input_signature
-    super(LayerError, self).__init__(self.message)
+    super().__init__(self.message)
 
   @property
   def message(self):
@@ -798,21 +802,16 @@ def _is_empty(container):
 
 def _find_frame(frame):
   """Find the frame with the caller on the stack."""
-  # TODO(lukaszkaiser): rewrite this function in a systematic way.
-  # We want to find the first place where the layer was called
-  # that is *not* an __init__ function of an inheriting layer.
-  # We also need to exclude a few decorator functions.
-  while frame.f_code.co_name in ['__init__', 'gin_wrapper', '_validate',
-                                 '_validate_forward_inputs', '_init']:
-    # We only skip __init__ in internal layers, return otherwise.
+  def _dirname_is_trax_layers_or_gin(frame):
+    """Skip frames coming from trax/layers or .../gin."""
     try:
-      dirname = frame.f_code.co_filename.split('/')[-2]
+      dirname1 = frame.f_code.co_filename.split('/')[-3]
+      dirname2 = frame.f_code.co_filename.split('/')[-2]
+      return (dirname1 == 'trax' and dirname2 == 'layers') or dirname2 == 'gin'
     except IndexError:
-      # Notebook cells have dummy filenames that do not contain any slashes
-      dirname = frame.f_code.co_filename
-    if dirname != 'layers' and frame.f_code.co_name == '__init__':
-      return frame
-    # If we are in an init, move up.
+      return False
+
+  while _dirname_is_trax_layers_or_gin(frame):
     frame = frame.f_back
   return frame
 
