@@ -52,7 +52,7 @@ class Serial(base.Layer):
   """
 
   def __init__(self, *sublayers, name=None, sublayers_to_print=None):
-    super(Serial, self).__init__(
+    super().__init__(
         name=name, sublayers_to_print=sublayers_to_print)
 
     sublayers = _ensure_flat(sublayers)
@@ -65,6 +65,7 @@ class Serial(base.Layer):
       self._state = tuple(None for l in sublayers)
 
   def forward(self, xs):
+    """Executes this layer as part of a forward pass through the model."""
     self._validate_forward_inputs(xs)
     state, weights = self.state, self.weights
     rngs = _split_rngs(self.rng, self._n_layers)
@@ -84,26 +85,27 @@ class Serial(base.Layer):
           f'number of sublayers ({n_layers}).')
 
     for layer, w, s, rng in zip(self.sublayers, weights, state, rngs):
-      inputs = _inputs_from_stack(layer, stack)
+      inputs = inputs_from_stack(stack, layer.n_in)
       outputs, s = layer.pure_fn(inputs, w, s, rng, use_cache=True)
-      stack = _outputs_onto_stack(layer, outputs, stack)
+      stack = outputs_onto_stack(outputs, stack, layer.n_in)
       new_state.append(s)
     self.state = new_state
     return stack
 
   # pylint: disable=protected-access
   def init_weights_and_state(self, input_signature):
+    """Initializes weights and state for inputs with the given signature."""
     weights = []
     states = []
     # In the code below, stack, inputs, and outputs are abstract (shapes and
     # dtypes), but weights and states are non-abstract actual values.
     stack = input_signature
     for sublayer in self.sublayers:
-      inputs = _inputs_from_stack(sublayer, stack)
+      inputs = inputs_from_stack(stack, sublayer.n_in)
       weights_or_cache_marker, state_or_cache_marker = (
           sublayer.init(inputs, use_cache=True))
       outputs, _ = sublayer._forward_abstract(inputs)
-      stack = _outputs_onto_stack(sublayer, outputs, stack)
+      stack = outputs_onto_stack(outputs, stack, sublayer.n_in)
 
       weights.append(weights_or_cache_marker)
       states.append(state_or_cache_marker)
@@ -172,7 +174,7 @@ class Parallel(base.Layer):
       A new layer in which each of the given sublayers applies to its
       corresponding span of elements in the dataflow stack.
     """
-    super(Parallel, self).__init__(name=name)
+    super().__init__(name=name)
     sublayers = self._validate(sublayers)
     self._n_layers = len(sublayers)
     self._sublayers = sublayers
@@ -182,6 +184,7 @@ class Parallel(base.Layer):
     self._state = tuple(None for l in sublayers)
 
   def forward(self, inputs):
+    """Executes this layer as part of a forward pass through the model."""
     n_layers, layers = self._n_layers, self.sublayers
     sublayer_inputs = self._allot_to_sublayers(inputs)
     state, weights = self.state, self.weights
@@ -217,6 +220,7 @@ class Parallel(base.Layer):
     return output
 
   def init_weights_and_state(self, input_signature):
+    """Initializes weights and state for inputs with the given signature."""
     sublayer_signatures = self._allot_to_sublayers(input_signature)
     inits = [layer.init(signature, use_cache=True)
              for layer, signature
@@ -270,15 +274,28 @@ class Parallel(base.Layer):
 
 
 class Concatenate(base.Layer):
-  """Concatenates n tensors into a single tensor."""
+  """Concatenates a number of tensors into a single tensor.
+
+  For example::
+
+      x = np.array([1, 2])
+      y = np.array([3, 4])
+      z = np.array([5, 6])
+      concat3 = tl.Concatenate(n_items=3)
+      z = concat3((x, y, z))  # z = [1, 2, 3, 4, 5, 6]
+
+  Use the `axis` argument to specify on which axis to concatenate the tensors.
+  By default it's the last axis, `axis=-1`, and `n_items=2`.
+  """
 
   def __init__(self, n_items=2, axis=-1):
     name = 'Concatenate' if axis == -1 else f'Concatenate_axis{axis}'
-    super(Concatenate, self).__init__(n_in=n_items, name=name)
+    super().__init__(n_in=n_items, name=name)
     self._n_items = n_items
     self._axis = axis
 
   def forward(self, xs):
+    """Executes this layer as part of a forward pass through the model."""
     return jnp.concatenate(xs, self._axis)
 
 
@@ -286,11 +303,12 @@ class Split(base.Layer):
   """Splits the input into n items along an axis."""
 
   def __init__(self, n_items=2, axis=-1):
-    super(Split, self).__init__(n_out=n_items)
+    super().__init__(n_out=n_items)
     self._n_items = n_items
     self._axis = axis
 
   def forward(self, inputs):
+    """Executes this layer as part of a forward pass through the model."""
     return tuple(jnp.split(inputs, self._n_items, self._axis))
 
 
@@ -368,7 +386,7 @@ class Scan(base.Layer):
   """
 
   def __init__(self, layer, axis=0, n_carry=1, remat=False):
-    super(Scan, self).__init__(n_in=layer.n_in, n_out=layer.n_out)
+    super().__init__(n_in=layer.n_in, n_out=layer.n_out)
     self._sublayers = [layer]
     self._n_carry = n_carry
     self._axis = axis
@@ -382,32 +400,35 @@ class Scan(base.Layer):
     return self._sublayers[0]
 
   def forward(self, inputs):
+    """Executes this layer as part of a forward pass through the model."""
     weights = self.weights[0]
     if isinstance(inputs, list):
       inputs = tuple(inputs)  # so that inputs structure matches outputs
     n_carry = self._n_carry
     def scannable_fn(x, carry_and_state):  # pylint: disable=invalid-name
-      carry, state = carry_and_state
+      carry, state, i = carry_and_state
       x_and_carry = x + carry if n_carry > 0 else x
+      rng = fastmath.random.fold_in(self.rng, i)
       res, new_state = self.sublayer.pure_fn(
-          x_and_carry, weights, state, self.rng, use_cache=True)
+          x_and_carry, weights, state, rng, use_cache=True)
       if n_carry > 0:
-        return (res[:-n_carry], (res[-n_carry:], new_state))
+        return (res[:-n_carry], (res[-n_carry:], new_state, i+1))
       else:
-        return (res, ([], new_state))
+        return (res, ([], new_state, i+1))
 
     if n_carry > 0:
       xs = inputs[:-n_carry]  # Split input stack into inputs and carry.
-      init = (inputs[-n_carry:], self.state[0])
+      init = (inputs[-n_carry:], self.state[0], jnp.array(0, dtype=jnp.int32))
     else:
-      xs, init = inputs, ([], self.state[0])
-    ys, (carry, new_state) = _scan(scannable_fn, xs, init,
-                                   axis=self._axis, remat=self._remat)
+      xs, init = inputs, ([], self.state[0], jnp.array(0, dtype=jnp.int32))
+    ys, (carry, new_state, _) = _scan(scannable_fn, xs, init,
+                                      axis=self._axis, remat=self._remat)
     res = ys + carry if n_carry > 0 else ys
     self.state = (new_state,)
     return res  # Put outputs and carry back on stack.
 
   def init_weights_and_state(self, input_signature):
+    """Initializes weights and state for inputs with the given signature."""
     n_carry = self._n_carry
     if n_carry == 0:
       if isinstance(input_signature, (list, tuple)):
@@ -431,7 +452,197 @@ class Scan(base.Layer):
       self.weights = (weights,)
 
 
+class Cond(base.Layer):
+  """Applies layers conditionally.
+
+  For parameters `cond`, `true`, and `false` runs the equivalent of `true(y)
+  if cond(x) else false(y)`, where `x` is `cond.n_in` elements from front of the
+  stack and `y` is the rest of the stack.
+  Exactly one of `true` and `false` functions is executed, so it can be used to
+  conditionally run long computations. The state of non-executed function is not
+  updated. Note that different branches may be executed on different devices
+  if `cond` returns different values on them.
+  By default 'false' function is an identity.
+
+  `cond` must return exactly one element: a Boolean value.
+  `true` and `false` must have the same n_in, and the same n_out.
+  """
+
+  def __init__(self, cond, true, false=None, name=None):
+    super(Cond, self).__init__(name=name)
+
+    if false is None:
+      self._identity_false_fun = True
+      # We don't need this function, but it will be useful for checking if
+      # 'true' has proper n_in/n_out.
+      false = Serial()
+      self._false = false
+    else:
+      self._identity_false_fun = False
+      self._false = false
+
+    sublayers = [cond, true, false]
+    self._sublayers = sublayers
+    self._n_layers = len(sublayers)
+    self._cond = cond
+    self._true = true
+
+    if cond.n_out != 1:
+      raise ValueError(
+          'cond.n_out must be 1: cond:{}->{}'.format(cond.n_in, cond.n_out))
+    if true.n_in != false.n_in:
+      raise ValueError(
+          'true.n_in and false.n_in must be equal: true:{}->{} ; false:{}->{}'
+          .format(true.n_in, true.n_out, false.n_in, false.n_out))
+    if true.n_out != false.n_out:
+      raise ValueError(
+          'true.n_out and false.n_out must be equal: true:{}->{} ; false:{}->{}'
+          .format(true.n_in, true.n_out, false.n_in, false.n_out))
+
+    self._n_in = cond.n_in + true.n_in
+    self._n_out = true.n_out
+    self._weights = tuple(None for l in sublayers)
+    self._state = tuple(None for l in sublayers)
+
+  # pylint: disable=protected-access
+  def init_weights_and_state(self, input_signature):
+    """Initializes weights and state for inputs with the given signature."""
+    weights = []
+    states = []
+    # In the code below, stack, inputs, and outputs are abstract (shapes and
+    # dtypes), but weights and states are non-abstract actual values.
+    stack = _make_tuple(input_signature)
+
+    # Inputs/outputs of `cond`.
+    inputs = inputs_from_stack(stack, self._cond.n_in)
+    weights_or_cache_marker, state_or_cache_marker = (
+        self._cond.init(inputs, use_cache=True))
+    weights.append(weights_or_cache_marker)
+    states.append(state_or_cache_marker)
+    self._cond._forward_abstract(inputs)
+    stack = _make_tuple(outputs_onto_stack([], stack, self._cond.n_in))
+
+    # Inputs/outputs of `true` and `false`.
+    for sublayer in [self._true, self._false]:
+      inputs = inputs_from_stack(stack, sublayer.n_in)
+      weights_or_cache_marker, state_or_cache_marker = (
+          sublayer.init(inputs, use_cache=True))
+      weights.append(weights_or_cache_marker)
+      states.append(state_or_cache_marker)
+
+    self.state = states
+    self.weights = weights
+    # pylint: enable=protected-access
+
+  def _validate_forward_inputs(self, xs):
+    xs = _make_tuple(xs)
+    if len(xs) < self.n_in:
+      raise ValueError(
+          f'Number of inputs ({len(xs)}) to Cond.forward less than n_in '
+          f'({self.n_in}).')
+
+  def forward(self, xs):
+    """Executes this layer as part of a forward pass through the model.
+
+    Args:
+      xs: Tensors of as required by the branches of this conditional.
+
+    Returns:
+      Tensors resulting from running the chosen branch.
+    """
+    # TODO(jaszczur): modify; it's a copy from SkippingSerial
+    self._validate_forward_inputs(xs)
+    layers_state = self.state
+    # Get 3 rngs, one for each layer.
+    rngs = _split_rngs(self.rng, 3)
+
+    # Prepare the stack and do some safety checks as in the parent class.
+    stack = _make_tuple(xs)
+    weights = self.weights
+    if len(weights) != 3:
+      raise ValueError('number of weights ({}) not equal to 3'
+                       .format(len(weights)))
+    if len(layers_state) != 3:
+      raise ValueError('length of state ({}) not equal to 3'
+                       .format(len(layers_state)))
+
+    def true_func(t):
+      outputs, new_true_state = self._true.pure_fn(
+          t[0][0], t[1][0], t[2][0], t[3][0])
+      # t[2][1] is old_false_state which is not changing if true is executed.
+      return outputs, (new_true_state, t[2][1])
+
+    def false_func(t):
+      if self._identity_false_fun:
+        # Memory optimization: we don't need pure_fn call.
+        return t[0][1], t[2]
+      outputs, new_false_state = self._false.pure_fn(
+          t[0][1], t[1][1], t[2][1], t[3][1])
+      # t[2][1] is old_true_state, which is not changing if false is executed.
+      return outputs, (t[2][0], new_false_state)
+
+    cond_inputs = inputs_from_stack(xs, self._cond.n_in)
+    cond_output, s = self._cond.pure_fn(cond_inputs, self.weights[0],
+                                        self.state[0], rngs[0], use_cache=True)
+    stack = outputs_onto_stack([], stack, self._cond.n_in)
+    self._cond.state = s
+
+    outputs, both_states = fastmath.cond(
+        cond_output,
+        true_func,
+        false_func,
+        [(stack, stack),
+         (self.weights[1], self.weights[2]),
+         (self.state[1], self.state[2]),
+         (rngs[1], rngs[2])]
+    )
+    stack = outputs_onto_stack([], stack, self._cond.n_in)
+
+    # We don't know which (`true` or `false`) branch was run, but both of them
+    # are adding (n_out) and removing (n_in) the same number of elements of the
+    # stack (this was checked in __init__). outputs_onto_stack just uses the
+    # layer's n_in, so we can pass either `true` or `false` to it.
+    # Note that `outputs` is the actual output of `true` or `false` branch,
+    # whichever was run, and we add it to the stack in any case.
+    stack = outputs_onto_stack(outputs, stack, self._true.n_in)
+    self._true.state = both_states[0]
+    self._false.state = both_states[1]
+    return _make_singleitem_or_original(stack)
+
+
 # pylint: disable=invalid-name
+def Chunk(layer, chunk_size, pass_unchunkable=True):
+  """Executes `layer` using batch chunks of size `chunk_size` to save memory."""
+  if chunk_size < 1:
+    return layer
+  def reshape_to_chunks(x):
+    chunk_batch = x.shape[0]
+    size = chunk_size
+    n_chunks = chunk_batch // size
+    if chunk_batch % size != 0:
+      if pass_unchunkable:
+        n_chunks = 1
+        size = chunk_batch
+      else:
+        raise ValueError(f'Chunk size {size} must divide batch '
+                         f'size {chunk_batch}')
+    return jnp.reshape(x, [n_chunks, size] + list(x.shape[1:]))
+  reshape_to_chunks_layer = base.PureLayer(
+      lambda xs: fastmath.nested_map(reshape_to_chunks, xs),
+      n_in=layer.n_in, n_out=layer.n_in, name='ReshapeToChunks')
+  def reshape_from_chunks(x):
+    batch_size = x.shape[0] * x.shape[1]
+    return jnp.reshape(x, [batch_size] + list(x.shape[2:]))
+  reshape_from_chunks_layer = base.PureLayer(
+      lambda xs: fastmath.nested_map(reshape_from_chunks, xs),
+      n_in=layer.n_out, n_out=layer.n_out, name='ReshapeFromChunks')
+  return Serial(
+      reshape_to_chunks_layer,
+      Scan(layer, axis=0, n_carry=0, remat=True),
+      reshape_from_chunks_layer,
+  )
+
+
 def Branch(*layers, name='Branch'):
   """Combinator that applies a list of layers in parallel to copies of inputs.
 
@@ -639,7 +850,7 @@ class Cache(base.Layer):
   """Applies a layer on the first run and returns the outputs on next calls."""
 
   def __init__(self, layer):
-    super(Cache, self).__init__(n_in=layer.n_in, n_out=layer.n_out)
+    super().__init__(n_in=layer.n_in, n_out=layer.n_out)
     self._sublayers = [layer]
 
   @property
@@ -661,11 +872,20 @@ class Cache(base.Layer):
     self.sublayer.state = state[1]
 
   def init_weights_and_state(self, input_signature):
+    """Initializes weights and state for inputs with the given signature."""
     weights, layer_state = self.sublayer.init(input_signature, use_cache=True)
     self.state = ((), layer_state)
     self._weights = (weights,)
 
   def forward(self, inputs):
+    """Executes this layer as part of a forward pass through the model.
+
+    Args:
+      inputs: Tensors required by the sublayer.
+
+    Returns:
+      Tensors resulting from running the sublayer the first time.
+    """
     state, weights = self.state, self.weights[0]
     if state[0] is ():  # pylint: disable=literal-comparison
       res, layer_state = self.sublayer.pure_fn(
@@ -689,7 +909,10 @@ class BatchLeadingAxes(base.Layer):
   """
 
   def __init__(self, layer, n_last_axes_to_keep=1):
-    super(BatchLeadingAxes, self).__init__(n_in=layer.n_in, n_out=layer.n_out)
+    if layer.n_in != 1 or layer.n_out != 1:
+      raise ValueError('BatchLeadingAxes currently only works for layers with '
+                       f'n_in = n_out = 1, got {(layer.n_in, layer.n_out)}.')
+    super().__init__(n_in=layer.n_in, n_out=layer.n_out)
     self._sublayers = [layer]
     self._n_last_axes_to_keep = n_last_axes_to_keep
     self._weights = (None,)
@@ -701,6 +924,7 @@ class BatchLeadingAxes(base.Layer):
     return self._sublayers[0]
 
   def forward(self, inputs):
+    """Executes this layer as part of a forward pass through the model."""
     batched_axes_shape = list(inputs.shape[:-self._n_last_axes_to_keep])
     batched_shape = [-1] + list(inputs.shape[-self._n_last_axes_to_keep:])
     inputs = jnp.reshape(inputs, batched_shape)
@@ -710,7 +934,14 @@ class BatchLeadingAxes(base.Layer):
     return jnp.reshape(res, batched_axes_shape + list(res.shape[1:]))
 
   def init_weights_and_state(self, input_signature):
-    weights, layer_state = self.sublayer.init(input_signature, use_cache=True)
+    """Initializes weights and state for inputs with the given signature."""
+    batched_size = 1
+    for d in input_signature.shape[:-self._n_last_axes_to_keep]:
+      batched_size *= d
+    batched_shape = [batched_size] + list(
+        input_signature.shape[-self._n_last_axes_to_keep:])
+    batched_signature = ShapeDtype(batched_shape, input_signature.dtype)
+    weights, layer_state = self.sublayer.init(batched_signature, use_cache=True)
     self.state = (layer_state,)
     self.weights = (weights,)
 
@@ -772,37 +1003,33 @@ def _split_rngs(rng, n_copies):
   return fastmath.random.split(rng, n_copies)
 
 
-def _inputs_from_stack(layer, stack, n_in=None):
-  """Returns the correct number/format of inputs for the given layer."""
-  is_stack_just_one_item = (_count_items(stack) == 1)
-  if isinstance(stack, (list, tuple)) and is_stack_just_one_item:
-    stack = stack[0]
-  if n_in is None:
-    n_in = layer.n_in
-  if n_in == 1 and is_stack_just_one_item:
-    return stack
-  elif n_in == 1:
-    return stack[0]
+def inputs_from_stack(stack, n):
+  """Returns n inputs from stack."""
+  stack = _make_tuple(stack)
+  return _make_singleitem_or_original(stack[:n])
+
+
+def outputs_onto_stack(outputs, stack, n):
+  """"Returns the new stack after removing n items and pushing outputs there."""
+  outputs = _make_tuple(outputs)
+  stack = _make_tuple(stack)
+  return _make_singleitem_or_original(outputs + stack[n:])
+
+
+def _make_tuple(xs):
+  """Returns a tuple from a list, a tuple, or a single element."""
+  if isinstance(xs, (list, tuple)):
+    return tuple(xs)
   else:
-    return stack[:n_in]
+    return (xs,)
 
 
-def _outputs_onto_stack(layer, outputs, stack, n_in=None, n_out=None):
-  """"Returns the new stack after outputs have been pushed onto it."""
-  if n_in is None:
-    n_in = layer.n_in
-  if n_out is None:
-    n_out = layer.n_out
-  if n_in < _count_items(stack):
-    if n_out == 1:
-      outputs = (outputs,)
-    return outputs + tuple(stack[n_in:])
+def _make_singleitem_or_original(xs):
+  """Returns a single element if possible, or the original list/tuple if not."""
+  if isinstance(xs, (list, tuple)) and len(xs) == 1:
+    return xs[0]
   else:
-    return outputs  # NOTE: can be single value or tuple.
-
-
-def _count_items(xs):
-  return len(xs) if isinstance(xs, (list, tuple)) else 1
+    return xs
 
 
 def _shape_without_axis(x, axis):
