@@ -31,6 +31,16 @@ def DivideBy(val):  # pylint: disable=invalid-name
   return tl.Fn('DivideBy', lambda x: x / val)
 
 
+def ReturnConst(val):  # pylint: disable=invalid-name
+  """Returns a simple const layer with n_in == 0 and n_out == 1."""
+  return tl.Fn('ReturnConst', lambda: val)
+
+
+def SmallerThan(val):  # pylint: disable=invalid-name
+  """Checks if the input is smaller than certain value."""
+  return tl.Fn('SmallerThan', lambda x: x < val)
+
+
 # TODO(jonni): Consider a more generic home for this utiliity function.
 def as_list(outputs):
   """Converts layer outputs to a nested list, for easier equality testing.
@@ -48,7 +58,7 @@ def as_list(outputs):
     testable using `assertEquals`.
   """
   if isinstance(outputs, (list, tuple)):
-    return [y.tolist() for y in outputs]
+    return [as_list(y) for y in outputs]
   else:
     return outputs.tolist()
 
@@ -77,6 +87,25 @@ class SerialTest(absltest.TestCase):
     y = layer(x)
     self.assertEqual(as_list(y), [1, 2, 3, 4])
 
+  def test_zero_in_one_out(self):
+    layer = tl.Serial(ReturnConst(np.array([3, 4, 5, 6])))
+    y = layer(())
+    self.assertEqual(as_list(y), [3, 4, 5, 6])
+
+  def test_one_in_two_out(self):
+    layer = tl.Serial(DivideBy(3),
+                      ReturnConst(np.array([3, 4, 5, 6])))
+    x = np.array([3, 6, 9, 12])
+    y = layer(x)
+    self.assertEqual(as_list(y), [[3, 4, 5, 6],
+                                  [1, 2, 3, 4]])
+
+  def test_const_div(self):
+    layer = tl.Serial(ReturnConst(np.array([3, 6, 9, 12])),
+                      DivideBy(3))
+    y = layer(())
+    self.assertEqual(as_list(y), [1, 2, 3, 4])
+
   def test_div_div(self):
     layer = tl.Serial(DivideBy(2.0), DivideBy(5.0))
     x = np.array([10, 20, 30])
@@ -455,6 +484,27 @@ class SwapTest(absltest.TestCase):
                                    [1, 2, 3]])
 
 
+class ChunkTest(absltest.TestCase):
+
+  def test_chunk(self):
+    layer = tl.Dense(4)
+    x = np.array([[1, 2, 3], [4, 5, 6]])
+    layer.init(x)
+    y = layer(x)
+    z = tl.Chunk(layer, 1)(x)
+    self.assertLess(np.sum((y - z)**2), 1e-5)  # y == z upto numerics
+
+  def test_chunk_uneven_numbers(self):
+    layer = tl.Dense(4)
+    x = np.array([[1, 2, 3], [4, 5, 6]])
+    layer.init(x)
+    y = layer(x)
+    z = tl.Chunk(layer, 3)(x)  # By default it should just pass
+    self.assertLess(np.sum((y - z)**2), 1e-5)  # y == z upto numerics
+    chunk_with_test = tl.Chunk(layer, 3, pass_unchunkable=False)
+    self.assertRaises(tl.LayerError, lambda: chunk_with_test(x))
+
+
 class SerialWithSideOutputsTest(absltest.TestCase):
 
   def test_serial_with_side_outputs_div_div(self):
@@ -469,7 +519,8 @@ class SerialWithSideOutputsTest(absltest.TestCase):
     self.assertEqual(output_shapes, [(3,), (5,), (2,)])
 
 
-@parameterized.named_parameters(('_' + b, b) for b in ('jax', 'tf'))
+@parameterized.named_parameters(
+    ('_' + b.value, b) for b in (fastmath.Backend.JAX, fastmath.Backend.TFNP))
 class ScanTest(parameterized.TestCase):
 
   def _AddWithCarry(self):  # pylint: disable=invalid-name
@@ -479,8 +530,8 @@ class ScanTest(parameterized.TestCase):
       return res, res  # output and carry are the same
     return tl.Fn('AddWithCarry', f, n_out=2)
 
-  def test_default_axis(self, backend_name):
-    with fastmath.use_backend(backend_name):
+  def test_default_axis(self, backend):
+    with fastmath.use_backend(backend):
       layer = tl.Scan(self._AddWithCarry())
       xs = [
           np.array([[0, 1, 2, 3],
@@ -497,8 +548,8 @@ class ScanTest(parameterized.TestCase):
                         [9000, 8111, 7222, 6333]
                        ])
 
-  def test_axis_1(self, backend_name):
-    with fastmath.use_backend(backend_name):
+  def test_axis_1(self, backend):
+    with fastmath.use_backend(backend):
       layer = tl.Scan(self._AddWithCarry(), axis=1)
       xs = [
           np.array([[0, 1, 2, 3],
@@ -519,13 +570,13 @@ class ScanTest(parameterized.TestCase):
                          7600]
                        ])
 
-  def test_multi_input(self, backend_name):
+  def test_multi_input(self, backend):
     def _MultiInputFn():  # pylint: disable=invalid-name
       def f(a, b, carry):
         return a + b, b, carry + 1
       return tl.Fn('MultiInputFn', f, n_out=2)
 
-    with fastmath.use_backend(backend_name):
+    with fastmath.use_backend(backend):
       layer = tl.Scan(_MultiInputFn(), axis=1)
       xs = [
           np.array([[0, 1, 2],
@@ -545,11 +596,11 @@ class ScanTest(parameterized.TestCase):
                          8003]
                        ])
 
-  def test_no_carry(self, backend_name):
+  def test_no_carry(self, backend):
     def _AddOne():  # pylint: disable=invalid-name
       return tl.Fn('AddOne', lambda x: x + 1)
 
-    with fastmath.use_backend(backend_name):
+    with fastmath.use_backend(backend):
       layer = tl.Scan(_AddOne(), n_carry=0)
       x = np.array([[1, 3, 7],
                     [10, 30, 70]])
@@ -558,6 +609,146 @@ class ScanTest(parameterized.TestCase):
                                     [11, 31, 71]])
 
 
+class CondTest(absltest.TestCase):
+
+  def test_basic_true(self):
+    cond = ReturnConst(True)
+    true = ReturnConst([2])
+    false = ReturnConst([5])
+    layer = tl.Cond(cond, true, false)
+    layer.init(())
+    xs = tuple()
+    ys = layer(xs)
+    self.assertEqual(as_list(ys), 2)
+
+  def test_basic_false(self):
+    cond = ReturnConst(False)
+    true = ReturnConst([2])
+    false = ReturnConst([5])
+    layer = tl.Cond(cond, true, false)
+    layer.init(())
+    xs = tuple()
+    ys = layer(xs)
+    self.assertEqual(as_list(ys), 5)
+
+  def test_complex_blocks(self):
+    cond = ReturnConst(True)
+    true = DivideBy(2.)
+    false = DivideBy(4.)
+    layer = tl.Cond(cond, true, false)
+    xs = [np.arange(5).astype(np.float32)]
+    layer.init(shapes.signature(xs))
+    ys = layer(xs)
+    self.assertEqual(as_list(ys), [0., 0.5, 1.0, 1.5, 2.0])
+
+  def test_condition_func_true(self):
+    cond = SmallerThan(3.0)
+    true = DivideBy(2.)
+    false = DivideBy(4.)
+    layer = tl.Cond(cond, true, false)
+    xs = (np.array(2.), np.array([4., 12.]))
+    layer.init(shapes.signature(xs))
+    ys = layer(xs)
+    self.assertEqual(as_list(ys), [2., 6.])
+
+  def test_condition_func_false(self):
+    cond = SmallerThan(3.0)
+    true = DivideBy(2.)
+    false = DivideBy(4.)
+    layer = tl.Cond(cond, true, false)
+    xs = (np.array(4.), np.array([4., 12.]))
+    layer.init(shapes.signature(xs))
+    ys = layer(xs)
+    self.assertEqual(as_list(ys), [1., 3.])
+
+  def test_condition_func_default_false(self):
+    cond = SmallerThan(3.0)
+    true = DivideBy(2.)
+    layer = tl.Cond(cond, true)
+    xs = (np.array(4.), np.array([4., 12.]))
+    layer.init(shapes.signature(xs))
+    ys = layer(xs)
+    self.assertEqual(as_list(ys), [4., 12.])
+
+  def test_exception_n_out(self):
+    cond = SmallerThan(3.0)
+    true = DivideBy(2.)
+    false = tl.Dup()
+    self.assertRaises(ValueError, lambda: tl.Cond(cond, true, false))
+
+  def test_exception_n_in(self):
+    cond = SmallerThan(3.0)
+    true = ReturnConst(2.)
+    false = DivideBy(2.)
+    self.assertRaises(ValueError, lambda: tl.Cond(cond, true, false))
+
+  def test_exception_run1(self):
+    # We expect exactly one input.
+    cond = SmallerThan(3.0)
+    true = ReturnConst(2.)
+    false = ReturnConst(5.)
+    def init_and_run(layer, xs):
+      layer.init(shapes.signature(xs))
+      layer(xs)
+    # It will pass with one input.
+    xs = np.array(4.)
+    layer = tl.Cond(cond, true, false)
+    init_and_run(layer, xs)
+    # It will fail with zero or two inputs.
+    for xs in ((), (np.array(4.), np.array([4., 12.]))):
+      layer = tl.Cond(cond, true, false)
+      # pylint: disable=cell-var-from-loop
+      self.assertRaises(Exception, lambda: init_and_run(layer, xs))
+
+  def test_exception_run2(self):
+    # We expect exactly two inputs.
+    cond = SmallerThan(3.0)
+    true = DivideBy(2.)
+    false = DivideBy(5.)
+    def init_and_run(layer, xs):
+      layer.init(shapes.signature(xs))
+      layer(xs)
+    # It will pass with two inputs.
+    xs = (np.array(4.), np.array([4., 12.]))
+    layer = tl.Cond(cond, true, false)
+    init_and_run(layer, xs)
+    # It will fail with zero or one input.
+    for xs in ((), (np.array(4.))):
+      # pylint: disable=cell-var-from-loop
+      self.assertRaises(Exception, lambda: init_and_run(layer, xs))
+
+  def test_weights_and_state(self):
+    cond = SmallerThan(3.0)
+    true = tl.Dense(5)
+    false = tl.Dense(5)
+    different = tl.Dense(5)
+    layer = tl.Cond(cond, true, false)
+    xs = (np.array(2.), np.array([0., 1., 2.]))
+    layer.init(shapes.signature(xs))
+
+    # weights
+    self.assertEqual(as_list(layer.weights),
+                     as_list((cond.weights, true.weights, false.weights)))
+    self.assertNotEqual(as_list(true.weights), as_list(false.weights))
+    self.assertNotEqual(as_list(true.weights), as_list(different.weights))
+
+    false.weights = true.weights
+    self.assertEqual(as_list(layer.weights),
+                     as_list((cond.weights, true.weights, true.weights)))
+
+    layer.weights = (cond.weights, true.weights, different.weights)
+    self.assertEqual(as_list(layer.weights),
+                     as_list((cond.weights, true.weights, different.weights)))
+    # state
+    self.assertEqual(as_list(layer.state),
+                     as_list((cond.state, true.state, false.state)))
+    # just check if simple assignments (setter from base.Layer) work correctly
+    # with Cond.init_weights_and_state ; all states are empty so there is no
+    # point in checking equality
+    false.state = true.state
+    layer.state = (cond.state, true.state, different.state)
+
+
 class BatchLeadingAxesTest(absltest.TestCase):
 
   def _Id3Dim(self):  # pylint: disable=invalid-name
@@ -565,7 +756,7 @@ class BatchLeadingAxesTest(absltest.TestCase):
     def f(x):
       assert len(x.shape) == 3
       return x
-    return tl.Fn('Id3Dim', f, n_out=2)
+    return tl.Fn('Id3Dim', f, n_out=1)
 
   def test_2axes(self):
     layer = tl.BatchLeadingAxes(self._Id3Dim(), n_last_axes_to_keep=2)
