@@ -20,7 +20,9 @@ from absl.testing import absltest
 import numpy as np
 
 from trax import shapes
+from trax.fastmath import numpy as jnp
 import trax.layers as tl
+import trax.layers.initializers as init
 
 
 class DenseTest(absltest.TestCase):
@@ -201,6 +203,106 @@ class DropoutTest(absltest.TestCase):
     self.assertEmpty(layer.weights)
 
 
+class WeightsTest(absltest.TestCase):
+  """Test Weights layer."""
+
+  def test_simple(self):
+    layer = tl.Weights(lambda shape, rng: jnp.zeros(shape, dtype=jnp.float32))
+    layer.init(())
+    y = layer(())
+    self.assertEqual(y.tolist(), 0.)
+
+  def test_shape(self):
+    layer = tl.Weights(init.RandomNormalInitializer(), (5, 10, 3))
+    layer.init(())
+    y = layer(())
+    self.assertEqual(y.shape, (5, 10, 3))
+
+  def test_simple_custom_initializer(self):
+    layer = tl.Weights(init.RandomNormalInitializer())
+    layer.init(())
+    y = layer(())
+    self.assertEqual(y.shape, ())
+    self.assertNotEqual(y.tolist(), 0.)
+
+  def test_custom_initializer_shape(self):
+    layer = tl.Weights(lambda shape, rng: jnp.zeros(shape, dtype=jnp.float32),
+                       (2, 2))
+    layer.init(())
+    y = layer(())
+    self.assertEqual(y.tolist(), [[0., 0.],
+                                  [0., 0.]])
+
+    layer = tl.Weights(init.RandomNormalInitializer(), (2, 2))
+    layer.init(())
+    y = layer(())
+    self.assertEqual(y.shape, (2, 2))
+    self.assertNotEqual(y.tolist(), [[0., 0.],
+                                     [0., 0.]])
+
+
+class SummaryScalarTest(absltest.TestCase):
+
+  def test_passes(self):
+    layer = tl.SummaryScalar('test')
+    x = np.array([[3., 5.], [2., 6.]])  # 10,000 values
+    y = layer(x)
+    self.assertEqual(y.tolist(), [[3., 5.], [2., 6.]])
+    self.assertEqual(layer.state['summary_test'].tolist(), 4.0)
+
+
+class RandomUniformTest(absltest.TestCase):
+  """Test Weights layer."""
+
+  def test_simple(self):
+    layer = tl.RandomUniform()
+    layer.init(())
+    y = layer(())
+    self.assertEqual(y.shape, ())
+    self.assertBetween(y, 0.0, 1.0)
+
+  def test_shape(self):
+    layer = tl.RandomUniform(shape=(5, 10, 3))
+    layer.init(())
+    y = layer(())
+    self.assertEqual(y.shape, (5, 10, 3))
+
+  def test_simple_range(self):
+    layer = tl.RandomUniform(1., 2., shape=(1000,))
+    layer.init(())
+    y = layer(())
+    self.assertEqual(y.shape, (1000,))
+    self.assertBetween(min(y.tolist()), 1., 2.)
+    self.assertBetween(max(y.tolist()), 1., 2.)
+    self.assertBetween(1.5, min(y.tolist()), max(y.tolist()))
+
+
+class LocallyConnected1dTest(absltest.TestCase):
+
+  def test_shape_kernel1(self):
+    for padding in ['WRAP', 'SAME', 'VALID']:
+      layer = tl.LocallyConnected1d(6, 1, padding=padding)
+      x = np.array([[0, 1], [2, 3], [4, 5]])
+      layer.init(shapes.signature(x))
+      y = layer(x)
+      self.assertEqual(y.shape, (3, 6))
+
+  def test_shape_kernel3(self):
+    for padding in ['WRAP', 'SAME']:
+      layer = tl.LocallyConnected1d(6, 3, padding=padding)
+      x = np.array([[0, 1], [2, 3], [4, 5]])
+      layer.init(shapes.signature(x))
+      y = layer(x)
+      self.assertEqual(y.shape, (3, 6))
+
+    for padding in ['VALID']:
+      layer = tl.LocallyConnected1d(6, 3, padding=padding)
+      x = np.array([[0, 1], [2, 3], [4, 5]])
+      layer.init(shapes.signature(x))
+      y = layer(x)
+      self.assertEqual(y.shape, (1, 6))
+
+
 class FlattenTest(absltest.TestCase):
 
   def test_keep_default(self):
@@ -253,5 +355,67 @@ class LogGaussianTest(absltest.TestCase):
     self.assertEqual(int(prob[1]), -6)
 
 
+class StopGradientTest(absltest.TestCase):
+
+  def test_passes(self):
+    layer = tl.StopGradient()
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, (2, 2))
+    self.assertEqual(y.tolist(), [[3., 5.], [2., 6.]])
+
+
+class MinMaxTest(absltest.TestCase):
+
+  def test_min(self):
+    layer = tl.Min()
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, (2,))
+    self.assertEqual(y.tolist(), [3., 2.])
+
+    layer = tl.Min(axis=0)
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, (2,))
+    self.assertEqual(y.tolist(), [2., 5.])
+
+    layer = tl.Min(axis=None)
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, ())
+    self.assertEqual(y.tolist(), 2.)
+
+    layer = tl.Min(keepdims=True)
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, (2, 1))
+    self.assertEqual(y.tolist(), [[3.], [2.]])
+
+  def test_max(self):
+    layer = tl.Max()
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, (2,))
+    self.assertEqual(y.tolist(), [5., 6.])
+
+    layer = tl.Max(axis=0)
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, (2,))
+    self.assertEqual(y.tolist(), [3., 6.])
+
+    layer = tl.Max(axis=None)
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, ())
+    self.assertEqual(y.tolist(), 6.)
+
+    layer = tl.Max(axis=0, keepdims=True)
+    x = np.array([[3., 5.], [2., 6.]])
+    y = layer(x)
+    self.assertEqual(y.shape, (1, 2))
+    self.assertEqual(y.tolist(), [[3., 6.]])
+
 if __name__ == '__main__':
   absltest.main()
