@@ -35,28 +35,86 @@ revealed several limitations, which this code attempts to address:
    only supported causal masking.
 """
 import functools
+import math
 import jax
 
 from trax import fastmath
 from trax.fastmath import numpy as np
 from trax.layers import base
+from trax.layers import initializers as init
 
 
 ####################################################### Functions
 
 
-def tie_in(x, y):
-  if fastmath.backend_name() == 'jax':
-    return jax.lax.tie_in(x, y)
-  return y
-
-
 def length_normalized(x, epsilon=1e-6):
   variance = np.mean(x**2, axis=-1, keepdims=True)
   norm_inputs = x / np.sqrt(variance + epsilon)
   return norm_inputs
 
 
+def hash_vecs(vecs, n_buckets_in, n_hashes, rng):
+  """Hash vectors into buckets.
+
+  Args:
+    vecs: vectors to hash, a tensor of shape [batch_size, depth]
+    n_buckets_in: an int or a list of ints, number of hash buckets;
+      if it is a list, we do hierarchical hashing as specified by the list
+    n_hashes: number of hashes
+    rng: random generator to use for hashing
+
+  Returns:
+    A pair (buckets, n_buckets) where buckets is a tensor of shape
+    [n_hashes, batch_size] of integers -- the hash bucket ids, and
+    n_buckets is an int, the total number of hash buckets, equal to
+    the product of all items in n_buckets_in.
+  """
+  # See https://arxiv.org/pdf/1509.02897.pdf
+  # We sample a different random rotation for each round of hashing to
+  # decrease the probability of hash misses.
+  if isinstance(n_buckets_in, int):
+    assert n_buckets_in % 2 == 0
+    rot_size = n_buckets_in
+    n_buckets = n_buckets_in
+  else:
+    # Factorize the hash if n_buckets_in is a list or tuple
+    rot_size, n_buckets = 0, 1
+    for factor in n_buckets_in:
+      assert factor % 2 == 0
+      rot_size += factor
+      n_buckets *= factor
+
+  rotations_shape = (vecs.shape[-1], n_hashes, rot_size // 2)
+  random_rotations = fastmath.random.normal(rng, rotations_shape).astype(
+      np.float32)
+  if fastmath.is_backend(fastmath.Backend.JAX):
+    rotated_vecs = np.einsum('tf,fhb->htb', vecs, random_rotations)
+  else:
+    random_rotations = np.reshape(random_rotations,
+                                  [-1, n_hashes * (rot_size // 2)])
+    rotated_vecs = np.dot(vecs, random_rotations)
+    rotated_vecs = np.reshape(rotated_vecs, [-1, n_hashes, rot_size//2])
+    rotated_vecs = np.transpose(rotated_vecs, (1, 0, 2))
+
+  if isinstance(n_buckets_in, int) or len(n_buckets_in) == 1:
+    rotated_vecs = np.concatenate([rotated_vecs, -rotated_vecs], axis=-1)
+    buckets = np.argmax(rotated_vecs, axis=-1).astype(np.int32)
+  else:
+    # Get the buckets for them and combine.
+    buckets, cur_sum, cur_product = None, 0, 1
+    for factor in n_buckets_in:
+      rv = rotated_vecs[..., cur_sum:cur_sum + (factor // 2)]
+      cur_sum += factor // 2
+      rv = np.concatenate([rv, -rv], axis=-1)
+      if buckets is None:
+        buckets = np.argmax(rv, axis=-1).astype(np.int32)
+      else:
+        buckets += cur_product * np.argmax(rv, axis=-1).astype(np.int32)
+      cur_product *= factor
+
+  return buckets, n_buckets  # buckets is now (n_hashes, batch_size)
+
+
 def look_adjacent(x, n_chunks_before, n_chunks_after):
   """Used to implement attention between consecutive chunks.
 
@@ -83,14 +141,16 @@ def look_adjacent(x, n_chunks_before, n_chunks_after):
 def mask_self_attention(
     dots, q_info, kv_info, causal=True, exclude_self=True, masked=False):
   """Performs masking for self-attention."""
+  q_info = q_info.astype(np.float32)
+  kv_info = kv_info.astype(np.float32)
   if causal:
-    mask = fastmath.lt(q_info, kv_info).astype(np.float32)
+    mask = fastmath.lt(q_info, kv_info)
     dots = dots - 1e9 * mask
   if exclude_self:
-    mask = np.equal(q_info, kv_info).astype(np.float32)
+    mask = np.equal(q_info, kv_info)
     dots = dots - 1e5 * mask
   if masked:
-    zeros_like_kv_info = tie_in(kv_info, np.zeros_like(kv_info))
+    zeros_like_kv_info = np.zeros_like(kv_info)
     mask = fastmath.lt(kv_info, zeros_like_kv_info).astype(np.float32)
     dots = dots - 1e9 * mask
   return dots
@@ -128,11 +188,18 @@ def attend(
   assert v is not None
   share_qk = (k is None)
 
+  # `q_info` and `kv_info` if supplied are 0 indexed, we want them to be 1
+  # indexed instead so that we can mask position 0 as well - see Github #820
+
   if q_info is None:
-    q_info = np.arange(q.shape[-2], dtype=np.int32)
+    q_info = np.arange(1, q.shape[-2] + 1, dtype=np.int32)
+  else:
+    q_info += 1
 
   if kv_info is None and not share_qk:
-    kv_info = np.arange(v.shape[-2], dtype=np.int32)
+    kv_info = np.arange(1, v.shape[-2] + 1, dtype=np.int32)
+  elif kv_info is not None:
+    kv_info += 1
 
   # Split q/k/v into chunks along the time axis, if desired.
   if q_chunk_len is not None:
@@ -185,9 +252,9 @@ def attend(
     # Dropout is broadcast across the bin dimension
     dropout_shape = (dots.shape[-2], dots.shape[-1])
     # TODO(kitaev): verify that tie-in is safe to remove (in light of jax fix)
-    keep_prob = tie_in(dots, 1.0 - dropout)
+    keep_prob = 1.0 - dropout
     keep = fastmath.random.bernoulli(rng, keep_prob, dropout_shape)
-    multiplier = keep.astype(dots.dtype) / tie_in(keep, keep_prob)
+    multiplier = keep.astype(dots.dtype) / keep_prob
     dots = dots * multiplier
 
   # The softmax normalizer (dots_logsumexp) is used by multi-round LSH attn.
@@ -201,45 +268,49 @@ def apply_broadcasted_dropout(vecs, dropout_rate, rng):
   """Apply dropout, broadcasted across all but the last dimension of `vecs`."""
   if dropout_rate > 0.0:
     assert rng is not None
-    keep_prob = tie_in(vecs, 1.0 - dropout_rate)
+    keep_prob = 1.0 - dropout_rate
     keep = fastmath.random.bernoulli(rng, keep_prob, (vecs.shape[-1],))
-    multiplier = keep.astype(vecs.dtype) / tie_in(keep, keep_prob)
+    multiplier = keep.astype(vecs.dtype) / keep_prob
     return vecs * multiplier
   else:
     return vecs
 
 
+# The new implementations below don't use custom_transforms in JAX but
+# do cause Tracer errors, so we don't use them for now.
+
+
 def permute_via_gather(val, permutation, inverse_permutation, axis=0):
   """Permutation helper for LSH attention."""
-  def permute_impl(val):
-    return np.take(val, permutation, axis=axis)
-  def permute_vjp(val):
-    permuted = permute_impl(fastmath.stop_gradient(val))
-    def vjpfun(permuted_grad):
-      # JAX autodiff would synthesize a scatter operation because it doesn't
-      # know that the indices are a permutatation. However on TPU, gathers are
-      # faster than scatters (at least in the regime the LSH attention uses).
-      return (np.take(permuted_grad, inverse_permutation, axis=axis),)
-    return permuted, vjpfun
-  permute = fastmath.custom_grad(permute_vjp, permute_impl)
-  return permute(val)
+  def permute_impl(p, unused_ip, val):
+    return np.take(val, p, axis=axis)
+  def permute_fwd(p, ip, val):
+    return np.take(val, p, axis=axis), ip
+  def permute_bwd(ip, permuted_grad):
+    # JAX autodiff would synthesize a scatter operation because it doesn't
+    # know that the indices are a permutation. However on TPU, gathers are
+    # faster than scatters (at least in the regime the LSH attention uses).
+    return (None, None, np.take(permuted_grad, ip, axis=axis))
+  permute = fastmath.custom_vjp(permute_impl, permute_fwd, permute_bwd)
+  return permute(permutation, inverse_permutation, val)
 
 
 def permute_via_sort(val, keys, inverse_keys, axis=0):
   """Permutation helper for LSH attention."""
-  def permute_impl(val):
+  def permute_impl(k, unused_ik, val):
     # On TPU, sorting scalars by key is faster than a gather.
-    _, permuted = fastmath.sort_key_val(keys, val, dimension=axis)
+    _, permuted = fastmath.sort_key_val(k, val, dimension=axis)
     return permuted
-  def permute_vjp(val):
-    permuted = permute_impl(fastmath.stop_gradient(val))
-    def vjpfun(permuted_grad):
-      _, val_grad = fastmath.sort_key_val(
-          inverse_keys, permuted_grad, dimension=axis)
-      return (val_grad,)
-    return permuted, vjpfun
-  permute = fastmath.custom_grad(permute_vjp, permute_impl)
-  return permute(val)
+  def permute_fwd(k, ik, val):
+    # On TPU, sorting scalars by key is faster than a gather.
+    _, permuted = fastmath.sort_key_val(k, val, dimension=axis)
+    return permuted, ik
+  def permute_bwd(ik, permuted_grad):
+    _, val_grad = fastmath.sort_key_val(
+        ik, permuted_grad, dimension=axis)
+    return (None, None, val_grad)
+  permute = fastmath.custom_vjp(permute_impl, permute_fwd, permute_bwd)
+  return permute(keys, inverse_keys, val)
 
 
 ####################################################### Classes
@@ -485,7 +556,7 @@ class EfficientAttentionBase(base.Layer):
              np.zeros_like(buf[:, :self.predict_drop_len])], axis=1)
 
       do_roll_mem = (mem_end + seqlen > self.predict_mem_len)
-      mem = jax.lax.cond(
+      mem = fastmath.cond(
           pred=do_roll_mem,
           true_operand=mem,
           true_fun=lambda x: fastmath.nested_map(roll_mem, x),
@@ -496,7 +567,7 @@ class EfficientAttentionBase(base.Layer):
       def update_mem(mem_element, new_vals):
         assert new_vals.shape[1] == seqlen
         if seqlen == 1:
-          return jax.ops.index_update(
+          return fastmath.index_update(
               mem_element, jax.ops.index[:, mem_end], new_vals[:, 0, ...])
         else:
           return jax.lax.dynamic_update_slice_in_dim(
@@ -526,7 +597,7 @@ class EfficientAttentionBase(base.Layer):
                        dtype=inp.dtype)
           ], axis=1)
         new_flat_mem.append(new_mem_val)
-      mem = jax.tree_unflatten(jax.tree_structure(mem), new_flat_mem)
+      mem, _ = fastmath.tree_unflatten(new_flat_mem, mem)
 
       # This code only works at the start of the sequence. There's no "assert"
       # primitive we can use to signal an error, so we instead signal the error
@@ -534,8 +605,9 @@ class EfficientAttentionBase(base.Layer):
       def replace_with_nan_if_not_seq_start(x):
         if x.dtype != np.float32:
           return x
-        return jax.lax.cond(
-            pred=jax.lax.eq(mem_end, 0), true_operand=x, true_fun=lambda x: x,
+        return fastmath.cond(
+            pred=np.equal(mem_end, np.array(0, dtype=mem_end.dtype)),
+            true_operand=x, true_fun=lambda x: x,
             false_operand=x, false_fun=lambda x: x * np.nan)
       inputs = fastmath.nested_map(replace_with_nan_if_not_seq_start, inputs)
       return inputs, state, 0, mem, np.minimum(seqlen, self.predict_mem_len)
@@ -653,12 +725,12 @@ class EfficientAttentionBase(base.Layer):
 
     def tree_update(tree, indices, new_values):
       return fastmath.nested_map_multiarg(
-          lambda x, y: jax.ops.index_update(x, jax.ops.index[indices], y),
+          lambda x, y: fastmath.index_update(x, jax.ops.index[indices], y),
           tree, new_values)
 
     def tree_add(tree, indices, new_values):
       return fastmath.nested_map_multiarg(
-          lambda x, y: jax.ops.index_add(x, jax.ops.index[indices], y),
+          lambda x, y: fastmath.index_add(x, jax.ops.index[indices], y),
           tree, new_values)
 
     if compute_grad:
@@ -684,7 +756,8 @@ class EfficientAttentionBase(base.Layer):
             leaves.append(non_differentiable_leaves.pop(0))
         assert not differentiable_leaves
         assert not non_differentiable_leaves
-        return jax.tree_unflatten(jax.tree_structure(inputs), leaves)
+        tree, _ = fastmath.tree_unflatten(leaves, inputs)
+        return tree
 
       def vjp(fn, inp, *args, has_aux=False):
         d_inp, nd_inp = split_differentiable(inp)
@@ -718,7 +791,7 @@ class EfficientAttentionBase(base.Layer):
           o_h, s_h = forward_fn(i_h, w_h)
 
         if compute_output:
-          o_all = jax.ops.index_add(o_all, example_idx, o_h)
+          o_all = fastmath.index_add(o_all, example_idx, o_h)
         if update_state:
           s_all = tree_update(s_all, idx, s_h)
         if compute_grad:
@@ -742,10 +815,10 @@ class EfficientAttentionBase(base.Layer):
         def forward_unbatched_h(i_h, w_h, s_h):
           return forward_unbatched(*i_h, weights=w_h, state=s_h)
         def forward_fn(i_mh, w_mh):
-          o_mh, new_s_mh = jax.vmap(
+          o_mh, new_s_mh = fastmath.vmap(
               forward_unbatched_h, in_axes=(None, 0, 0), out_axes=0)(
                   i_mh, w_mh, s_mh)
-          o_mh = o_mh.sum(0)
+          o_mh = np.sum(o_mh, axis=0)
           return o_mh, new_s_mh
 
         if compute_grad:
@@ -757,7 +830,7 @@ class EfficientAttentionBase(base.Layer):
           o_mh, s_mh = forward_fn(i_mh, w_mh)
 
         if compute_output:
-          o_all = jax.ops.index_add(o_all, example_idx, o_mh)
+          o_all = fastmath.index_add(o_all, example_idx, o_mh)
         if update_state:
           s_all = tree_update(s_all, state_range, s_mh)
         if compute_grad:
@@ -769,10 +842,10 @@ class EfficientAttentionBase(base.Layer):
       def forward_single_example(i_x, w_all, s_x):
         def forward_unbatched_h(i_h, w_h, s_h):
           return forward_unbatched(*i_h, weights=w_h, state=s_h)
-        o_x, s_x = jax.vmap(
+        o_x, s_x = fastmath.vmap(
             forward_unbatched_h, in_axes=(None, 0, 0), out_axes=(0, 0))(
                 i_x, w_all, s_x)
-        o_x = o_x.sum(0)
+        o_x = np.sum(o_x, axis=0)
         return o_x, s_x
       def run_inner(idx, loop_val):
         """Runs one slice of attention (all heads for one or more examples)."""
@@ -789,25 +862,26 @@ class EfficientAttentionBase(base.Layer):
                                  (-1, self.n_heads) + s.shape[1:]),
             state)
         def forward_fn(i_mex, w_all):
-          o_mex, new_s_mex = jax.vmap(
+          o_mex, new_s_mex = fastmath.vmap(
               forward_single_example, in_axes=(0, None, 0), out_axes=(0, 0))(
                   i_mex, w_all, s_mex)
           new_s_mex = fastmath.nested_map(
               lambda s: np.reshape(s, (n_parallel_heads,) + s.shape[2:]),
               new_s_mex)
-          return o_mex, new_s_mex
+          return o_mex.astype(i_mex[0].dtype), new_s_mex
 
         if compute_grad:
           o_mex, backward_fn, s_mex = vjp(forward_fn, i_mex, weights,
                                           has_aux=True)
           ct_mex = output_grad[example_range]
-          assert o_mex.shape == ct_mex.shape
+          assert o_mex.shape == ct_mex.shape, str(ct_mex.shape)
+          assert o_mex.dtype == ct_mex.dtype, str(ct_mex.dtype)
           i_ct_mex, w_ct_mex = backward_fn(ct_mex)
         else:
           o_mex, s_mex = forward_fn(i_mex, weights)
 
         if compute_output:
-          o_all = jax.ops.index_add(o_all, jax.ops.index[example_range], o_mex)
+          o_all = fastmath.index_add(o_all, jax.ops.index[example_range], o_mex)
         if update_state:
           s_all = tree_update(s_all, state_range, s_mex)
         if compute_grad:
@@ -836,7 +910,7 @@ class EfficientAttentionBase(base.Layer):
       for idx in range(loop_hi):
         loop_val = run_inner(idx, loop_val)
     else:
-      loop_val = jax.lax.fori_loop(
+      loop_val = fastmath.fori_loop(
           0, loop_hi, run_inner, loop_val)
 
     (o_all, s_all, i_ct_all, w_ct_all) = loop_val
@@ -1001,12 +1075,12 @@ class SelfAttention(EfficientAttentionBase):
     mask_fn = functools.partial(
         mask_self_attention,
         causal=self.causal, exclude_self=self.share_qk, masked=self.masked)
-    q_info = kv_info = tie_in(x, np.arange(q.shape[-2], dtype=np.int32))
+    q_info = kv_info = np.arange(q.shape[-2], dtype=np.int32)
 
     assert (mask is not None) == self.masked
     if self.masked:
       # mask is a boolean array (True means "is valid token")
-      ones_like_mask = tie_in(x, np.ones_like(mask, dtype=np.int32))
+      ones_like_mask = np.ones_like(mask, dtype=np.int32)
       kv_info = kv_info * np.where(mask, ones_like_mask, -ones_like_mask)
 
     o, _ = attend(
@@ -1033,7 +1107,7 @@ class SelfAttention(EfficientAttentionBase):
     else:
       w_q, w_k, w_v, w_o = weights
 
-    q_range = q_start + tie_in(x, np.arange(q_len, dtype=np.int32))
+    q_range = q_start + np.arange(q_len, dtype=np.int32)
     if q_len == 1:
       # On TPU, np.matmul(a[:1], b) and np.matmul(a, b)[:1] are not
       # floating-point equivalent, at least in non-jitted code. We correct the
@@ -1053,7 +1127,7 @@ class SelfAttention(EfficientAttentionBase):
         mask_self_attention,
         causal=self.causal, exclude_self=self.share_qk, masked=self.masked)
     q_info = q_range
-    kv_info = tie_in(x, np.arange(k.shape[-2], dtype=np.int32))
+    kv_info = np.arange(k.shape[-2], dtype=np.int32)
 
     if self.chunk_len is not None and q_len > self.chunk_len:
       assert q_start == 0
@@ -1088,9 +1162,9 @@ class LSHSelfAttention(SelfAttention):
                n_heads=2, d_qk=64, d_v=64, share_qk='unused',
                causal=False,
                masked=False,
-               chunk_len=None, n_chunks_before=1, n_chunks_after=0,
+               chunk_len=128, n_chunks_before=1, n_chunks_after=0,
                n_hashes=1,
-               n_buckets=256,
+               n_buckets=None,
                mode='train',
                predict_mem_len=2048, predict_drop_len=256,
                attention_dropout=0.0,
@@ -1139,60 +1213,32 @@ class LSHSelfAttention(SelfAttention):
       return (buckets, buckets_idx, rng)
 
   def hash_vectors(self, vecs, rng, mask=None):
-    # See https://arxiv.org/pdf/1509.02897.pdf
-    # We sample a different random rotation for each round of hashing to
-    # decrease the probability of hash misses.
-    if isinstance(self.n_buckets, int):
-      assert self.n_buckets % 2 == 0
-      rot_size = self.n_buckets
-      n_buckets = self.n_buckets
-    else:
-      # Factorize the hash if self.n_buckets is a list or tuple
-      rot_size, n_buckets = 0, 1
-      for factor in self.n_buckets:
-        assert factor % 2 == 0
-        rot_size += factor
-        n_buckets *= factor
+    n_buckets_list = self.n_buckets
 
-    rotations_shape = (vecs.shape[-1], self.n_hashes, rot_size // 2)
-    rng = fastmath.stop_gradient(tie_in(vecs, rng))
-    random_rotations = fastmath.random.normal(rng, rotations_shape).astype(
-        np.float32)
-    if fastmath.backend_name() == 'jax':
-      rotated_vecs = np.einsum('tf,fhb->htb', vecs, random_rotations)
-    else:
-      random_rotations = np.reshape(random_rotations,
-                                    [-1, self.n_hashes * (rot_size // 2)])
-      rotated_vecs = np.dot(vecs, random_rotations)
-      rotated_vecs = np.reshape(rotated_vecs, [-1, self.n_hashes, rot_size//2])
-      rotated_vecs = np.transpose(rotated_vecs, (1, 0, 2))
+    # Determine the number of buckets needed from input length if not set.
+    if n_buckets_list is None:
+      length = vecs.shape[0]
+      n_buckets = 2 * max(1, length // self.chunk_len)
+      if n_buckets <= 128:
+        n_buckets_list = n_buckets
+      else:  # Factorize n_buckets.
+        n_buckets_div = 2**math.ceil(math.log2(math.sqrt(n_buckets)))
+        # Both factors must be even.
+        n_buckets_rest = 2 * (n_buckets // (2 * n_buckets_div))
+        n_buckets_list = [n_buckets_div, n_buckets_rest]
 
-    if isinstance(self.n_buckets, int) or len(self.n_buckets) == 1:
-      rotated_vecs = np.concatenate([rotated_vecs, -rotated_vecs], axis=-1)
-      buckets = np.argmax(rotated_vecs, axis=-1).astype(np.int32)
-    else:
-      # Get the buckets for them and combine.
-      buckets, cur_sum, cur_product = None, 0, 1
-      for factor in self.n_buckets:
-        rv = rotated_vecs[..., cur_sum:cur_sum + (factor // 2)]
-        cur_sum += factor // 2
-        rv = np.concatenate([rv, -rv], axis=-1)
-        if buckets is None:
-          buckets = np.argmax(rv, axis=-1).astype(np.int32)
-        else:
-          buckets += cur_product * np.argmax(rv, axis=-1).astype(np.int32)
-        cur_product *= factor
+    # Hash vectors.
+    buckets, n_buckets = hash_vecs(vecs, n_buckets_list, self.n_hashes, rng)
 
     if mask is not None:
       n_buckets += 1  # Create an extra bucket for padding tokens only
       buckets = np.where(mask[None, :], buckets, n_buckets - 1)
 
-    # buckets is now (self.n_hashes, seqlen). Next we add offsets so that
+    # buckets is now (n_hashes, seqlen). Next we add offsets so that
     # bucket numbers from different hashing rounds don't overlap.
-    offsets = tie_in(buckets, np.arange(self.n_hashes, dtype=np.int32))
+    offsets = np.arange(self.n_hashes, dtype=np.int32)
     offsets = np.reshape(offsets * n_buckets, (-1, 1))
     buckets = np.reshape(buckets + offsets, (-1,))
-
     return buckets
 
   def forward_unbatched(self, x, mask=None, *, weights, state, rng,
@@ -1223,7 +1269,7 @@ class LSHSelfAttention(SelfAttention):
     seqlen = x.shape[0]
     assert int(buckets.shape[0]) == self.n_hashes * seqlen
 
-    ticker = tie_in(x, np.arange(self.n_hashes * seqlen, dtype=np.int32))
+    ticker = np.arange(self.n_hashes * seqlen, dtype=np.int32)
     buckets_and_t = seqlen * buckets + (ticker % seqlen)
     buckets_and_t = fastmath.stop_gradient(buckets_and_t)
 
@@ -1248,7 +1294,7 @@ class LSHSelfAttention(SelfAttention):
     if self.masked:
       # mask is a boolean array (True means "is valid token")
       smask = np.take(mask, st, axis=0)
-      ones_like_mask = tie_in(x, np.ones_like(smask, dtype=np.int32))
+      ones_like_mask = np.ones_like(smask, dtype=np.int32)
       kv_info = q_info * np.where(smask, ones_like_mask, -ones_like_mask)
 
     so, slogits = attend(
@@ -1281,7 +1327,9 @@ class LSHSelfAttention(SelfAttention):
                                     weights, state, rng, update_state):
     assert update_state, (
         'This setting not supported (e.g. no backprop for fast inference)')
-    if isinstance(q_start, int) and q_start == 0 and q_len > 1:
+    if q_len > 1:
+      if isinstance(q_start, int):
+        assert q_start == 0, 'Chunks larger than 1 only work at start for now.'
       if x.shape[0] % self.chunk_len == 0:
         x_padded = x
       else:
@@ -1322,7 +1370,7 @@ class LSHSelfAttention(SelfAttention):
       new_buckets = np.reshape(new_buckets, (-1,))
       return new_buckets
 
-    buckets = jax.lax.cond(
+    buckets = fastmath.cond(
         pred=buckets_idx > q_start,
         true_operand=buckets,
         true_fun=roll_buckets,
@@ -1333,7 +1381,7 @@ class LSHSelfAttention(SelfAttention):
     attend_rng, output_rng = fastmath.random.split(rng)
     w_q, w_v, w_o = weights
 
-    q_range = q_start + tie_in(x, np.arange(q_len, dtype=np.int32))
+    q_range = q_start + np.arange(q_len, dtype=np.int32)
     # On TPU, np.matmul(a[:1], b) and np.matmul(a, b)[:1] are not
     # floating-point equivalent, at least in non-jitted code. We correct the
     # discrepancy by duplicating the slice. Floating-point noise may not be
@@ -1369,7 +1417,8 @@ class LSHSelfAttention(SelfAttention):
     mask_fn = functools.partial(
         mask_self_attention, causal=True, masked=True, exclude_self=True)
     q_info = q_start + np.arange(q_len, dtype=np.int32)
-    kv_info = kv_indices
+    kv_info = kv_indices.astype(np.int32)
+    q_info = q_info.astype(np.int32)
     # TODO(kitaev): is it better to mask out attention across buckets?
     # kv_info = np.where(is_valid_target[kv_indices], kv_indices, -kv_indices)
     o, _ = attend(
@@ -1469,3 +1518,91 @@ class EncDecAttention(EfficientAttentionBase):
     out = np.matmul(o, w_o)
     out = apply_broadcasted_dropout(out, self.output_dropout, output_rng)
     return out, state
+
+
+class LSHFF(base.Layer):
+  """Feed-forward block with LSH.
+
+  The original (non-LSH) feed-forward block is a triple Dense(d_ff)-Relu-Dense
+  that takes an input, makes it of size d_ff (usually larger than it was) and
+  then brings it back to the original size after Relu. It is commonly used in
+  Transformer models where it often accounts for most of the trainable weights.
+
+  The original block can be slow in decoding due to the need to fetch a lot of
+  weights from memory. The LSH block aims to exploit this sparsity. So in the
+  first Dense(d_ff) layer, instead of making a full matrix multiplication,
+  this block only multiplies by the parts of the weights matrix that have
+  the highest chance to give non-0 after Relu. This is determined by taking
+  a number of locality-sensitive hashes and masking to only include weights
+  that have one hash identical to the multiplied element.
+  """
+
+  def __init__(self, d_ff, n_buckets, n_hashes=4, mode='train',
+               kernel_initializer=init.GlorotUniformInitializer(),
+               bias_initializer=init.RandomNormalInitializer(1e-6)):
+    """Returns a LSH feed-forward block."""
+    super().__init__(name=f'LSHFF_{d_ff}')
+    self._mode = mode
+    self._d_ff = d_ff
+    self._n_buckets = n_buckets
+    self._n_hashes = n_hashes
+    self._kernel_initializer = kernel_initializer
+    self._bias_initializer = bias_initializer
+
+  def forward(self, x):
+    """Executes this layer as part of a forward pass through the model.
+
+    Args:
+      x: Tensor of same shape and dtype as the input signature used to
+          initialize this layer.
+
+    Returns:
+      Tensor of same shape and dtype as the input.
+    """
+    w1, w2, b2 = self.weights
+    x_shape = x.shape
+    x = np.reshape(x, [-1, x_shape[-1]])  # Easier to operate on flattened x.
+
+    # Hash x into hash buckets; x_buckets is [n_hashes, joint_batch].
+    x_buckets, _ = hash_vecs(x, self._n_buckets, self._n_hashes, self.rng)
+
+    # Hash w1 into hash buckets; w1_buckets is [n_hashes, d_ff].
+    # Note that we use the same self.rng - so the same hash vectors as for x.
+    w1_buckets, _ = hash_vecs(w1, self._n_buckets, self._n_hashes, self.rng)
+
+    # Create a mask to determine which x's have the same hash as which w1's.
+    # First: just subtract the hashes and make them non-negative.
+    hash_mask = (x_buckets[:, :, None] - w1_buckets[:, None, :])**2
+    hash_mask = fastmath.stop_gradient(hash_mask)  # make sure no gradients here
+    # hash_mask is [n_hashes, joint_batch, d_ff], 0 iff hashes were equal
+    hash_mask = 1 - np.minimum(hash_mask, 1)  # now 1 if equal, 0 otherwise
+    # we now sum over n_hashes and use min, it's 1 iff any of n_hashes was equal
+    hash_mask = np.minimum(np.sum(hash_mask, axis=0), 1)
+    hash_mask = hash_mask.astype(np.float32)  # convert to float to use mask
+
+    # First dense layer of the block, with hash masking.
+    mid = np.dot(x, w1.T) * hash_mask  # [joint_batch, d_ff]
+
+    # Relu and the second dense layer, as in a standard feed-forward block.
+    # Note: we merge the second block into this layer because of future plans,
+    # not anything implemented yet. The potential gain would be as follows:
+    # in predict mode, we would pre-hash (once) both w1 and w2 and only do
+    # matmuls (and memory copies) for the parts that correspond to the hash
+    # of the input. The hash of w1 determines which parts of Relu are 0, so
+    # it also determines which parts of w2 can be skipped.
+    relu = np.where(mid <= 0, np.zeros_like(mid), mid)
+    res = np.dot(relu, w2) + b2
+    return np.reshape(res, x_shape)  # un-flatten if needed
+
+  def init_weights_and_state(self, input_signature):
+    """Randomly initializes this layer's weights."""
+    d_model = input_signature.shape[-1]
+    shape_w1 = (self._d_ff, d_model)
+    shape_w2 = (self._d_ff, d_model)
+    shape_b2 = (d_model,)
+
+    rng_w1, rng_w2, rng_b2 = fastmath.random.split(self.rng, 3)
+    w1 = self._kernel_initializer(shape_w1, rng_w1)
+    w2 = self._kernel_initializer(shape_w2, rng_w2)
+    b2 = self._bias_initializer(shape_b2, rng_b2)
+    self.weights = (w1, w2, b2)
