@@ -33,7 +33,7 @@ class AxialPositionalEncoding(layer_base.Layer):
   def __init__(self, shape=(64, 64, 3), d_embs=(384, 384, 256),
                kernel_initializer=init.RandomNormalInitializer(1.0),
                dropout=0.0, dropout_broadcast_dims=(), mode='train'):
-    super(AxialPositionalEncoding, self).__init__()
+    super().__init__()
     self._kernel_initializer = kernel_initializer
     assert len(shape) == len(d_embs)
     self._shape = shape
@@ -78,9 +78,6 @@ class AxialPositionalEncoding(layer_base.Layer):
       for dim in self._dropout_broadcast_dims:
         noise_shape[dim] = 1
       keep_prob = 1.0 - self._dropout
-      if fastmath.backend_name() == 'jax':
-        keep_prob = jax.lax.tie_in(
-            inputs, jnp.full((), keep_prob, dtype=inputs.dtype))
       keep = fastmath.random.bernoulli(rng, keep_prob, tuple(noise_shape))
       multiplier = keep.astype(inputs.dtype) / keep_prob
       return inputs + jnp.reshape(emb * multiplier, inputs.shape)
@@ -113,7 +110,7 @@ class FixedBasePositionalEncoding(layer_base.Layer):
   def __init__(self, bases=[11, 13, 14, 15], n_digits=8,  #  pylint: disable=dangerous-default-value
                start_from_zero_one_in=100, base_dropout_one_in=100,
                mode='train', initializer=init.RandomUniformInitializer(1e-4)):
-    super(FixedBasePositionalEncoding, self).__init__()
+    super().__init__()
     self._bases = bases
     self._n_digits = n_digits
     self._mode = mode
@@ -130,10 +127,11 @@ class FixedBasePositionalEncoding(layer_base.Layer):
     positions = jnp.arange(0, length)[None, :]
     if self._mode == 'train':
       # In 1% of training cases still start from 0 to be exactly as in eval.
-      start_from_nonzero = jax.random.randint(
+      start_from_nonzero = fastmath.random.randint(
           rng1, (batch_size,), 0, self._start_from_zero_one_in)
       start_from_nonzero = jnp.minimum(1, start_from_nonzero)
-      random_start = jax.random.randint(rng2, (batch_size,), 0, max_pos-length)
+      random_start = fastmath.random.randint(
+          rng2, (batch_size,), 0, max_pos-length)
       random_start *= start_from_nonzero
       positions += random_start[:, None]
     res = []
@@ -147,7 +145,7 @@ class FixedBasePositionalEncoding(layer_base.Layer):
         pos_embeddings.append(cur_indices.astype(jnp.float32)[:, :, None] * s)
       embeddings = jnp.concatenate(pos_embeddings, axis=-1)
       if self._mode == 'train':
-        base_dropout = jax.random.randint(
+        base_dropout = fastmath.random.randint(
             rng3, (batch_size,), 0, self._base_dropout_one_in)
         base_dropout = jnp.minimum(1, base_dropout).astype(jnp.float32)
         embeddings *= base_dropout[:, None, None]
