@@ -15,18 +15,15 @@
 
 """End to end test for Reformer."""
 
-import contextlib
 import os
-import tempfile
 
 from absl.testing import absltest
 import gin
-from tensorflow.compat.v1 import test
-from tensorflow.compat.v1.io import gfile
 
-import trax
+from trax import test_utils
 from trax.models.reformer import reformer  # pylint: disable=unused-import
 from trax.supervised import trainer_lib
+from trax.tf_numpy import numpy as tf_np  # pylint: disable=unused-import
 
 pkg_dir, _ = os.path.split(__file__)
 _TESTDATA = os.path.join(pkg_dir, 'testdata')
@@ -39,16 +36,9 @@ class ReformerE2ETest(absltest.TestCase):
     super().setUp()
     gin.clear_config()
     gin.add_config_file_search_path(_CONFIG_DIR)
-
-  @contextlib.contextmanager
-  def tmp_dir(self):
-    tmp = tempfile.mkdtemp(dir=test.get_temp_dir())
-    yield tmp
-    gfile.rmtree(tmp)
+    test_utils.ensure_flag('test_tmpdir')
 
   def test_reformer_wmt_ende(self):
-    trax.fastmath.disable_jit()
-
     batch_size_per_device = 2
     steps = 1
     n_layers = 2
@@ -63,29 +53,66 @@ class ReformerE2ETest(absltest.TestCase):
     gin.bind_parameter('Reformer.n_decoder_layers', n_layers)
     gin.bind_parameter('Reformer.d_ff', d_ff)
 
-    with self.tmp_dir() as output_dir:
-      _ = trainer_lib.train(output_dir=output_dir)
+    output_dir = self.create_tempdir().full_path
+    _ = trainer_lib.train(output_dir=output_dir)
 
-  def test_reformer_noencdecattn_wmt_ende(self):
-    trax.fastmath.disable_jit()
-
-    batch_size_per_device = 1  # Ignored, but needs to be set.
+  def test_reformer2_wmt_ende(self):
+    batch_size_per_device = 2
     steps = 1
     n_layers = 2
     d_ff = 32
 
-    gin.parse_config_file('reformer_noencdecattn_wmt_ende.gin')
+    gin.parse_config_file('reformer2_wmt_ende.gin')
 
     gin.bind_parameter('data_streams.data_dir', _TESTDATA)
     gin.bind_parameter('batcher.batch_size_per_device', batch_size_per_device)
-    gin.bind_parameter('batcher.buckets', ([512], [1, 1]))  # batch size 1.
+    gin.bind_parameter('batcher.buckets',
+                       ([512], [batch_size_per_device, batch_size_per_device]))
     gin.bind_parameter('train.steps', steps)
-    gin.bind_parameter('ReformerNoEncDecAttention.n_encoder_layers', n_layers)
-    gin.bind_parameter('ReformerNoEncDecAttention.n_decoder_layers', n_layers)
-    gin.bind_parameter('ReformerNoEncDecAttention.d_ff', d_ff)
+    gin.bind_parameter('Reformer2.n_encoder_layers', n_layers)
+    gin.bind_parameter('Reformer2.n_decoder_layers', n_layers)
+    gin.bind_parameter('Reformer2.d_ff', d_ff)
+
+    output_dir = self.create_tempdir().full_path
+    _ = trainer_lib.train(output_dir=output_dir)
+
+  def test_reformer2_copy(self):
+    batch_size_per_device = 2
+    steps = 1
+    n_layers = 2
+    d_ff = 32
+
+    gin.parse_config_file('reformer2_copy.gin')
+
+    gin.bind_parameter('batcher.batch_size_per_device', batch_size_per_device)
+    gin.bind_parameter('batcher.buckets', ([64], [1, 1]))  # batch size 1.
+    gin.bind_parameter('train.steps', steps)
+    gin.bind_parameter('Reformer2.n_encoder_layers', n_layers)
+    gin.bind_parameter('Reformer2.n_decoder_layers', n_layers)
+    gin.bind_parameter('Reformer2.d_ff', d_ff)
+
+    output_dir = self.create_tempdir().full_path
+    _ = trainer_lib.train(output_dir=output_dir)
+
+  def test_reformer_copy(self):
+    batch_size_per_device = 2
+    steps = 1
+    n_layers = 2
+    d_ff = 32
+    d_model = 32
+
+    gin.parse_config_file('reformer_copy.gin')
+
+    gin.bind_parameter('data_streams.data_dir', _TESTDATA)
+    gin.bind_parameter('batcher.batch_size_per_device', batch_size_per_device)
+    gin.bind_parameter('train.steps', steps)
+    gin.bind_parameter('ReformerLM.n_layers', n_layers)
+    gin.bind_parameter('ReformerLM.d_ff', d_ff)
+    gin.bind_parameter('ReformerLM.d_model', d_model)
+
+    output_dir = self.create_tempdir().full_path
+    _ = trainer_lib.train(output_dir=output_dir)
 
-    with self.tmp_dir() as output_dir:
-      _ = trainer_lib.train(output_dir=output_dir)
 
 if __name__ == '__main__':
   absltest.main()
