@@ -18,6 +18,7 @@
 
 from trax import layers as tl
 from trax import models
+from trax.fastmath import numpy as jnp
 
 
 def _Batch(x, batch_axes):
@@ -65,7 +66,9 @@ def Value(
     batch_axes=None,
     mode='train',
     is_discrete=False,
-    vocab_size=2
+    vocab_size=2,
+    multiplicative_action_injection=False,
+    head_init_range=None,
 ):
   """Attaches a value head to a model body."""
   if body is None:
@@ -76,19 +79,24 @@ def Value(
   def ActionInjector(mode):
     if inject_actions:
       if is_discrete:
-        encode_layer = tl.Parallel(
-            tl.Dense(inject_actions_dim),
-            tl.Embedding(vocab_size, inject_actions_dim)
+        action_encoder = tl.Embedding(vocab_size, inject_actions_dim)
+      else:
+        action_encoder = tl.Dense(inject_actions_dim)
+      encoders = tl.Parallel(
+          tl.Dense(inject_actions_dim),
+          action_encoder,
+      )
+      if multiplicative_action_injection:
+        action_injector = tl.Serial(
+            tl.Fn('TanhMulGate', lambda x, a: x * jnp.tanh(a)),
+            tl.LayerNorm()  # compensate for reduced variance
         )
       else:
-        encode_layer = tl.Parallel(
-            tl.Dense(inject_actions_dim),
-            tl.Dense(inject_actions_dim),
-        )
+        action_injector = tl.Add()
       return tl.Serial(
           # Input: (body output, actions).
-          encode_layer,
-          tl.Add(),
+          encoders,
+          action_injector,
           models.PureMLP(
               layer_widths=(inject_actions_dim,) * inject_actions_n_layers,
               out_activation=True,
@@ -99,11 +107,17 @@ def Value(
     else:
       return []
 
+  head_kwargs = {}
+  if head_init_range is not None:
+    head_kwargs['kernel_initializer'] = tl.RandomUniformInitializer(
+        lim=head_init_range
+    )
+
   return tl.Serial(
       _Batch(normalizer(mode=mode), batch_axes),
       _Batch(body(mode=mode), batch_axes),
       ActionInjector(mode=mode),
-      tl.Dense(1),
+      tl.Dense(1, **head_kwargs),
   )
 
 
@@ -136,3 +150,31 @@ def PolicyAndValue(
           value_top(mode=mode),
       ),
   )
+
+
+def Quality(
+    body=None,
+    normalizer=None,
+    batch_axes=None,
+    mode='train',
+    n_actions=2,
+    head_init_range=None,
+):
+  """The network takes as input an observation and outputs values of actions."""
+
+  if body is None:
+    body = lambda mode: []
+  if normalizer is None:
+    normalizer = lambda mode: []
+
+  head_kwargs = {}
+  if head_init_range is not None:
+    head_kwargs['kernel_initializer'] = tl.RandomUniformInitializer(
+        lim=head_init_range
+    )
+
+  return tl.Serial(
+      _Batch(normalizer(mode=mode), batch_axes),
+      _Batch(body(mode=mode), batch_axes),
+      tl.Dense(n_actions, **head_kwargs),
+  )
