@@ -24,6 +24,9 @@ from trax.optimizers import base
 from trax.optimizers import momentum
 from trax.optimizers import rms_prop
 from trax.optimizers import sm3
+from trax.optimizers import trainer
+from trax.optimizers.trainer import ReversibleSerialTrainer
+from trax.optimizers.trainer import Trainer
 
 
 def opt_configure(*args, **kwargs):
