@@ -21,7 +21,7 @@ from trax.optimizers import base as opt_base
 
 
 class Adafactor(opt_base.Optimizer):
-  """Adafactor optimizer."""
+  """Adafactor optimizer, as described in https://arxiv.org/abs/1804.04235."""
 
   def __init__(self,
                learning_rate=0.05,
@@ -62,7 +62,7 @@ class Adafactor(opt_base.Optimizer):
     self._do_clipping = do_clipping
     self._do_momentum = do_momentum
     # Dynamically configurable parameters will be passed to the update function.
-    super(Adafactor, self).__init__(
+    super().__init__(
         learning_rate=learning_rate,
         beta1=beta1,
         decay_rate=decay_rate,
