@@ -18,7 +18,6 @@
 
 from trax import fastmath
 from trax.fastmath import numpy as jnp
-from trax.layers import base as layers
 
 
 class Optimizer(object):
@@ -162,7 +161,7 @@ class Optimizer(object):
 
   def _l2_norm(self, flat_list):
     """Returns the aggregate L2 norm of a list of tensors."""
-    if fastmath.backend_name() == 'jax':
+    if fastmath.is_backend(fastmath.Backend.JAX):
       norm = jnp.sqrt(sum(jnp.vdot(x, x) for x in flat_list))
     else:  # TODO(lukaszkaiser): add vdot to TF-numpy
       norm = jnp.sqrt(sum(jnp.sum(x*x) for x in flat_list))
@@ -213,4 +212,4 @@ def clip_grads(grad_tree, max_norm):
   """Clip gradients stored as a pytree of arrays to maximum norm `max_norm`."""
   norm = l2_norm(grad_tree)
   normalize = lambda g: jnp.where(norm < max_norm, g, g * (max_norm / norm))
-  return layers.nested_map(grad_tree, normalize)
+  return fastmath.nested_map(grad_tree, normalize)
