@@ -34,7 +34,7 @@ from trax.rl import training as rl_training
 from trax.supervised import lr_schedules as lr
 
 
-class ActorCriticTrainer(rl_training.PolicyTrainer):
+class ActorCriticAgent(rl_training.PolicyAgent):
   """Trains policy and value models using actor-critic methods.
 
   Attrs:
@@ -57,9 +57,11 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
                n_replay_epochs=1,
                scale_value_targets=False,
                q_value=False,
-               q_value_aggregate_max=True,
+               q_value_aggregate='max',
+               q_value_temperature=1.0,
                q_value_n_samples=1,
-               **kwargs):  # Arguments of PolicyTrainer come here.
+               q_value_normalization=False,
+               **kwargs):  # Arguments of PolicyAgent come here.
     """Configures the actor-critic trainer.
 
     Args:
@@ -70,8 +72,9 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
       value_batch_size: Batch size for value model training.
       value_train_steps_per_epoch: Number of steps are we using to train the
           value model in each epoch.
-      value_evals_per_epoch: Number of value trainer evaluations per RL epoch;
-          only affects metric reporting.
+      value_evals_per_epoch: Number of value trainer evaluations per RL epoch.
+          Every evaluation, we also synchronize the weights of the target
+          network.
       value_eval_steps: Number of value trainer steps per evaluation; only
           affects metric reporting.
       n_shared_layers: Number of layers to share between value and policy
@@ -86,10 +89,15 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
       scale_value_targets: If `True`, scale value function targets by
           `1 / (1 - gamma)`.
       q_value: If `True`, use Q-values as baselines.
-      q_value_aggregate_max: If `True`, aggregate Q-values with max (or mean).
+      q_value_aggregate: How to aggregate Q-values. Options: 'mean', 'max',
+        'softmax', 'logsumexp'.
+      q_value_temperature: Temperature parameter for the 'softmax' and
+        'logsumexp' aggregation methods.
       q_value_n_samples: Number of samples to average over when calculating
           baselines based on Q-values.
-      **kwargs: Arguments for `PolicyTrainer` superclass.
+      q_value_normalization: How to normalize Q-values before aggregation.
+          Allowed values: 'std', 'abs', `None`. If `None`, don't normalize.
+      **kwargs: Arguments for `PolicyAgent` superclass.
     """
     self._n_shared_layers = n_shared_layers
     self._value_batch_size = value_batch_size
@@ -98,7 +106,7 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
     self._value_eval_steps = value_eval_steps
 
     # The 2 below will be initalized in super.__init__ anyway, but are needed
-    # to construct value batches which are needed before PolicyTrainer init
+    # to construct value batches which are needed before PolicyAgent init
     # since policy input creation calls the value model -- hence this code.
     self._task = task
     self._max_slice_length = kwargs.get('max_slice_length', 1)
@@ -112,8 +120,10 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
       self._value_network_scale = 1
 
     self._q_value = q_value
-    self._q_value_aggregate_max = q_value_aggregate_max
+    self._q_value_aggregate = q_value_aggregate
+    self._q_value_temperature = q_value_temperature
     self._q_value_n_samples = q_value_n_samples
+    self._q_value_normalization = q_value_normalization
 
     is_discrete = isinstance(self._task.action_space, gym.spaces.Discrete)
     self._is_discrete = is_discrete
@@ -138,7 +148,7 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
         self._value_eval_model.pure_fn, fastmath.device_count(), do_mean=False)
 
     # Initialize policy training.
-    super(ActorCriticTrainer, self).__init__(task, **kwargs)
+    super().__init__(task, **kwargs)
 
     # Initialize training of the value function.
     value_output_dir = kwargs.get('output_dir', None)
@@ -156,7 +166,16 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
         loss_fn=tl.L2Loss(),
         inputs=self._value_inputs,
         output_dir=value_output_dir,
-        metrics={'value_loss': tl.L2Loss()})
+        metrics={'value_loss': tl.L2Loss(),
+                 'value_mean': self.value_mean})
+
+  @property
+  def value_mean(self):
+    """The mean value of the value function."""
+    # TODO(henrykm): A better solution would take into account the masks
+    def f(values):
+      return jnp.mean(values)
+    return tl.Fn('ValueMean', f)
 
   @property
   def _value_model_signature(self):
@@ -221,14 +240,44 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
     values = jnp.squeeze(values, axis=-1)  # Remove the singleton depth dim.
     return (values, actions, log_probs)
 
-  def _aggregate_values(self, values, aggregate_max, act_log_probs):
+  def _aggregate_values(self, values, aggregate, act_log_probs):
+    # Normalize the Q-values before aggragetion, so it can adapt to the scale
+    # of the returns. This does not affect mean and max aggregation.
+    scale = 1
+    epsilon = 1e-5
+    if self._q_value_normalization == 'std':
+      scale = jnp.std(values) + epsilon
+    elif self._q_value_normalization == 'abs':
+      scale = jnp.mean(jnp.abs(values - jnp.mean(values))) + epsilon
+    values /= scale
+
+    temp = self._q_value_temperature
     if self._q_value:
-      if aggregate_max:
+      assert values.shape[:2] == (
+          self._value_batch_size, self._q_value_n_samples
+      )
+      if aggregate == 'max':
+        # max_a Q(s, a)
         values = jnp.max(values, axis=1)
-      elif self._sample_all_discrete_actions:
-        values = jnp.sum(values * jnp.exp(act_log_probs), axis=1)
+      elif aggregate == 'softmax':
+        # sum_a (Q(s, a) * w(s, a))
+        # where w(s, .) = softmax (Q(s, .) / T)
+        weights = tl.Softmax(axis=1)(values / temp)
+        values = jnp.sum(values * weights, axis=1)
+      elif aggregate == 'logsumexp':
+        # log(mean_a exp(Q(s, a) / T)) * T
+        n = values.shape[1]
+        values = (fastmath.logsumexp(values / temp, axis=1) - jnp.log(n)) * temp
       else:
-        values = jnp.mean(values, axis=1)
+        assert aggregate == 'mean'
+        # mean_a Q(s, a)
+        if self._sample_all_discrete_actions:
+          values = jnp.sum(values * jnp.exp(act_log_probs), axis=1)
+        else:
+          values = jnp.mean(values, axis=1)
+
+    # Re-scale the Q-values after aggregation.
+    values *= scale
     return np.array(values)  # Move the values to CPU.
 
   def value_batches_stream(self):
@@ -245,7 +294,7 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
           np_trajectory.observations, np_trajectory.dist_inputs
       )
       values = self._aggregate_values(
-          values, self._q_value_aggregate_max, act_log_probs)
+          values, self._q_value_aggregate, act_log_probs)
 
       # TODO(pkozakowski): Add some shape assertions and docs.
       # Calculate targets based on the advantages over the target network - this
@@ -255,8 +304,6 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
           returns=np_trajectory.returns,
           values=values,
           dones=np_trajectory.dones,
-          gamma=self._task.gamma,
-          n_extra_steps=self._added_policy_slice_length,
       )
       length = advantages.shape[1]
       values = values[:, :length]
@@ -304,7 +351,7 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
         include_final_state=False):
       (values, _, act_log_probs) = self._run_value_model(
           np_trajectory.observations, np_trajectory.dist_inputs)
-      values = self._aggregate_values(values, False, act_log_probs)
+      values = self._aggregate_values(values, 'mean', act_log_probs)
       if len(values.shape) != 2:
         raise ValueError('Values are expected to have shape ' +
                          '[batch_size, length], got: %s' % str(values.shape))
@@ -338,6 +385,9 @@ class ActorCriticTrainer(rl_training.PolicyTrainer):
           self._value_train_steps_per_epoch // self._value_evals_per_epoch,
           self._value_eval_steps,
       )
+      # Update the target value network.
+      self._value_eval_model.weights = self._value_trainer.model_weights
+      self._value_eval_model.state = self._value_trainer.model_state
 
     # Copy value weights and state to policy trainer.
     if self._n_shared_layers > 0:
@@ -405,7 +455,7 @@ def _copy_model_weights_and_state(  # pylint: disable=invalid-name
 ### Implementations of common actor-critic algorithms.
 
 
-class AdvantageBasedActorCriticTrainer(ActorCriticTrainer):
+class AdvantageBasedActorCriticAgent(ActorCriticAgent):
   """Base class for advantage-based actor-critic algorithms."""
 
   def __init__(
@@ -414,12 +464,17 @@ class AdvantageBasedActorCriticTrainer(ActorCriticTrainer):
       advantage_estimator=rl_advantages.td_lambda,
       advantage_normalization=True,
       advantage_normalization_epsilon=1e-5,
+      added_policy_slice_length=0,
       **kwargs
   ):
-    self._advantage_estimator = advantage_estimator
+    self._advantage_estimator = advantage_estimator(
+        gamma=task.gamma, margin=added_policy_slice_length
+    )
     self._advantage_normalization = advantage_normalization
     self._advantage_normalization_epsilon = advantage_normalization_epsilon
-    super(AdvantageBasedActorCriticTrainer, self).__init__(task, **kwargs)
+    super().__init__(
+        task, added_policy_slice_length=added_policy_slice_length, **kwargs
+    )
 
   def policy_inputs(self, trajectory, values):
     """Create inputs to policy model from a TrajectoryNp and values."""
@@ -430,8 +485,6 @@ class AdvantageBasedActorCriticTrainer(ActorCriticTrainer):
         returns=trajectory.returns,
         values=values,
         dones=trajectory.dones,
-        gamma=self._task.gamma,
-        n_extra_steps=self._added_policy_slice_length,
     )
     # Observations should be the same length as advantages - so if we are
     # using n_extra_steps, we need to trim the length to match.
@@ -495,7 +548,7 @@ class AdvantageBasedActorCriticTrainer(ActorCriticTrainer):
 
   @property
   def policy_metrics(self):
-    metrics = super(AdvantageBasedActorCriticTrainer, self).policy_metrics
+    metrics = super().policy_metrics
     metrics.update({
         'advantage_mean': self.advantage_mean,
         'advantage_std': self.advantage_std,
@@ -519,15 +572,15 @@ class AdvantageBasedActorCriticTrainer(ActorCriticTrainer):
     ])
 
 
-class A2CTrainer(AdvantageBasedActorCriticTrainer):
-  """Trains policy and value models using the A2C algortithm."""
+class A2C(AdvantageBasedActorCriticAgent):
+  """Trains policy and value models using the A2C algorithm."""
 
   on_policy = True
 
   def __init__(self, task, entropy_coeff=0.01, **kwargs):
     """Configures the A2C Trainer."""
     self._entropy_coeff = entropy_coeff
-    super(A2CTrainer, self).__init__(task, **kwargs)
+    super().__init__(task, **kwargs)
 
   @property
   def policy_loss_given_log_probs(self):
@@ -561,10 +614,10 @@ class A2CTrainer(AdvantageBasedActorCriticTrainer):
     return tl.Fn('A2CLoss', f)
 
 
-class PPOTrainer(AdvantageBasedActorCriticTrainer):
+class PPO(AdvantageBasedActorCriticAgent):
   """The Proximal Policy Optimization Algorithm aka PPO.
 
-  Trains policy and value models using the PPO algortithm.
+  Trains policy and value models using the PPO algorithm.
   """
 
   on_policy = True
@@ -573,7 +626,7 @@ class PPOTrainer(AdvantageBasedActorCriticTrainer):
     """Configures the PPO Trainer."""
     self._entropy_coeff = entropy_coeff
     self._epsilon = epsilon
-    super(PPOTrainer, self).__init__(task, **kwargs)
+    super().__init__(task, **kwargs)
 
   @property
   def policy_loss_given_log_probs(self):
@@ -596,7 +649,7 @@ class PPOTrainer(AdvantageBasedActorCriticTrainer):
                          ', %s != %s' % (new_log_probs.shape, mask.shape))
 
       # The ratio between new_probs and old_probs expressed
-      # using log_probs and exponentaion
+      # using log_probs and exponentiation
       probs_ratio = jnp.exp(new_log_probs - old_log_probs)
       if advantages.shape != probs_ratio.shape:
         raise ValueError('New log-probs and old log probs shapes '
@@ -670,7 +723,7 @@ def AWRLoss(beta, w_max):  # pylint: disable=invalid-name
   return tl.Fn('AWRLoss', f)
 
 
-class AWRTrainer(AdvantageBasedActorCriticTrainer):
+class AWR(AdvantageBasedActorCriticAgent):
   """Trains policy and value models using AWR."""
 
   on_policy = False
@@ -679,7 +732,7 @@ class AWRTrainer(AdvantageBasedActorCriticTrainer):
     """Configures the AWR Trainer."""
     self._beta = beta
     self._w_max = w_max
-    super(AWRTrainer, self).__init__(task, **kwargs)
+    super().__init__(task, **kwargs)
 
   @property
   def policy_loss_given_log_probs(self):
@@ -688,7 +741,7 @@ class AWRTrainer(AdvantageBasedActorCriticTrainer):
 
   @property
   def policy_metrics(self):
-    metrics = super(AWRTrainer, self).policy_metrics
+    metrics = super().policy_metrics
     metrics.update(awr_metrics(self._beta))
     return metrics
 
@@ -705,7 +758,7 @@ def SamplingAWRLoss(beta, w_max, reweight=False, sampled_all_discrete=False):  #
   return tl.Fn('SamplingAWRLoss', f)
 
 
-class SamplingAWRTrainer(AdvantageBasedActorCriticTrainer):
+class SamplingAWR(AdvantageBasedActorCriticAgent):
   """Trains policy and value models using Sampling AWR."""
 
   on_policy = False
@@ -715,7 +768,7 @@ class SamplingAWRTrainer(AdvantageBasedActorCriticTrainer):
     self._beta = beta
     self._w_max = w_max
     self._reweight = reweight
-    super(SamplingAWRTrainer, self).__init__(task, q_value=True, **kwargs)
+    super().__init__(task, q_value=True, **kwargs)
 
   def _policy_inputs_to_advantages(self, preprocess):
     """A layer that computes advantages from policy inputs."""
