@@ -31,7 +31,7 @@ from trax.supervised import lr_schedules as lr
 
 
 # pylint: disable=g-long-lambda
-class ActorCriticJointTrainer(rl_training.RLTrainer):
+class ActorCriticJointAgent(rl_training.Agent):
   """Trains a joint policy-and-value model using actor-critic methods."""
 
   def __init__(self, task,
@@ -68,7 +68,7 @@ class ActorCriticJointTrainer(rl_training.RLTrainer):
       n_replay_epochs: how many last epochs to take into the replay buffer;
            > 1 only makes sense for off-policy algorithms.
     """
-    super(ActorCriticJointTrainer, self).__init__(
+    super().__init__(
         task,
         n_trajectories_per_epoch=n_trajectories_per_epoch,
         output_dir=output_dir,
@@ -206,7 +206,7 @@ class ActorCriticJointTrainer(rl_training.RLTrainer):
           self._supervised_eval_steps)
 
 
-class PPOJointTrainer(ActorCriticJointTrainer):
+class PPOJoint(ActorCriticJointAgent):
   """The Proximal Policy Optimization Algorithm aka PPO.
 
   Trains policy and value models using the PPO algortithm.
@@ -222,7 +222,7 @@ class PPOJointTrainer(ActorCriticJointTrainer):
     self._epsilon = epsilon
     self._value_loss_coeff = value_loss_coeff
     self._entropy_coeff = entropy_coeff
-    super(PPOJointTrainer, self).__init__(task, **kwargs)
+    super().__init__(task, **kwargs)
     self._trainer = supervised.Trainer(
         model=self._joint_model,
         optimizer=self._optimizer,
@@ -455,7 +455,7 @@ class PPOJointTrainer(ActorCriticJointTrainer):
     return tl.Fn('PPOObjectiveMean', f)
 
 
-class A2CJointTrainer(ActorCriticJointTrainer):
+class A2CJoint(ActorCriticJointAgent):
   """The A2C algorithm.
 
   Trains policy and value models using the A2C algortithm.
@@ -468,7 +468,7 @@ class A2CJointTrainer(ActorCriticJointTrainer):
     """Configures the A2C Trainer."""
     self._value_loss_coeff = value_loss_coeff
     self._entropy_coeff = entropy_coeff
-    super(A2CJointTrainer, self).__init__(task, **kwargs)
+    super().__init__(task, **kwargs)
     self._trainer = supervised.Trainer(
         model=self._joint_model,
         optimizer=self._optimizer,
@@ -618,7 +618,7 @@ class A2CJointTrainer(ActorCriticJointTrainer):
     return tl.Fn('A2CObjectiveMean', f, n_out=1)
 
 
-class AWRJointTrainer(ActorCriticJointTrainer):
+class AWRJoint(ActorCriticJointAgent):
   """Trains a joint policy-and-value model using AWR."""
 
   # TODO(henrykm): value_loss_coeff looks like a common parameter
@@ -628,7 +628,7 @@ class AWRJointTrainer(ActorCriticJointTrainer):
     self._beta = beta
     self._w_max = w_max
     self._value_loss_coeff = value_loss_coeff
-    super(AWRJointTrainer, self).__init__(task, **kwargs)
+    super().__init__(task, **kwargs)
 
   def batches_stream(self):
     """Use the RLTask self._task to create inputs to the value model."""
