@@ -276,6 +276,7 @@ class RLTask:
                initial_trajectories=1,
                gamma=0.99,
                dm_suite=False,
+               random_starts=True,
                max_steps=None,
                timestep_to_np=None,
                num_stacked_frames=1,
@@ -292,6 +293,7 @@ class RLTask:
         trajectories are stored.
       gamma: float: discount factor for calculating returns.
       dm_suite: whether we are using the DeepMind suite or the gym interface
+      random_starts: use random starts for training of Atari agents.
       max_steps: Optional int: stop all trajectories at that many steps.
       timestep_to_np: a function that turns a timestep into a numpy array
         (ie., a tensor); if None, we just use the state of the timestep to
@@ -303,20 +305,48 @@ class RLTask:
     if isinstance(env, str):
       self._env_name = env
       if dm_suite:
-        env = environments.load_from_settings(
+        eval_env = environments.load_from_settings(
             platform='atari',
             settings={
                 'levelName': env,
                 'interleaved_pixels': True,
-                'zero_indexed_actions': True
+                'zero_indexed_actions': True,
+                'phase': 'testing',
             })
-        env = atari_wrapper.AtariWrapper(environment=env,
-                                         num_stacked_frames=num_stacked_frames)
+        eval_env = atari_wrapper.AtariWrapper(
+            environment=eval_env,
+            max_abs_reward=None,
+            num_stacked_frames=num_stacked_frames)
+        if random_starts:
+          env = environments.load_from_settings(
+              platform='atari',
+              settings={
+                  'levelName': env,
+                  'interleaved_pixels': True,
+                  'zero_indexed_actions': True,
+                  'phase': 'training'
+              })
+        else:
+          # This is the default behavior used so far
+          # in QWR experiments (implicitly 'condition': 'at30').
+          env = environments.load_from_settings(
+              platform='atari',
+              settings={
+                  'levelName': env,
+                  'interleaved_pixels': True,
+                  'zero_indexed_actions': True
+              })
+        env = atari_wrapper.AtariWrapper(
+            environment=env,
+            num_stacked_frames=num_stacked_frames)
       else:
         env = gym.make(env)
+        eval_env = env
     else:
       self._env_name = type(env).__name__
+      eval_env = env
     self._env = env
+    self._eval_env = eval_env
     self._dm_suite = dm_suite
     self._max_steps = max_steps
     self._gamma = gamma
@@ -448,7 +478,7 @@ class RLTask:
   def save_to_file(self, file_name):
     """Save this task to file."""
     # Save trajectories from new epochs first.
-    epochs_to_save = [e for e in self._trajectories.keys()
+    epochs_to_save = [e for e in self._trajectories
                       if e not in self._saved_epochs_unchanged]
     for epoch in epochs_to_save:
       training.pickle_to_file(self._trajectories[epoch],
@@ -463,11 +493,14 @@ class RLTask:
                   'all_epochs': list(self._trajectories.keys())}
     training.pickle_to_file(dictionary, file_name, gzip=False)
 
-  def play(self, policy, max_steps=None):
+  def play(self, policy, max_steps=None, only_eval=False):
     """Play an episode in env taking actions according to the given policy."""
     if max_steps is None:
       max_steps = self._max_steps
-    cur_trajectory = play(self._env, policy, self._dm_suite, max_steps)
+    if only_eval:
+      cur_trajectory = play(self._eval_env, policy, self._dm_suite, max_steps)
+    else:
+      cur_trajectory = play(self._env, policy, self._dm_suite, max_steps)
     cur_trajectory.calculate_returns(self._gamma)
     return cur_trajectory
 
@@ -482,7 +515,8 @@ class RLTask:
     """Collect experience in env playing the given policy."""
     max_steps = max_steps or self.max_steps
     if n_trajectories:
-      new_trajectories = [self.play(policy, max_steps=max_steps)
+      new_trajectories = [self.play(policy, max_steps=max_steps,
+                                    only_eval=only_eval)
                           for _ in range(n_trajectories)]
     elif n_interactions:
       # TODO(pkozakowski): Test this mode of experience collection.
@@ -538,7 +572,7 @@ class RLTask:
 
   def remove_epoch(self, epoch):
     """Useful when we need to remove an unwanted trajectory."""
-    if epoch in self._trajectories.keys():
+    if epoch in self._trajectories:
       self._trajectories.pop(epoch)
 
   def trajectory_stream(self, epochs=None, max_slice_length=None,
@@ -697,7 +731,7 @@ class RLTask:
         ])
         # Where act, rew and ret will usually have the following shape:
         # [batch_size, trajectory_length-1], which we call [B, L-1].
-        # Observations are more complex and will usuall be [B, L] + S where S
+        # Observations are more complex and will usually be [B, L] + S where S
         # is the shape of the observation space (self.observation_space.shape).
         # We stop the recursion at level 1, so we pass lists of arrays into
         # pad().
