@@ -86,7 +86,7 @@ class Trainer(object):
     self._is_chief, _, self._n_devices, rng = (
         training.init_host_and_devices(n_devices, random_seed))
     self._should_save_checkpoints = should_save_checkpoints and self._is_chief
-    self._checkpoints_at = checkpoints_at or []
+    self._checkpoints_at = checkpoints_at if checkpoints_at is not None else []
     self._should_write_summaries = should_write_summaries
     if not output_dir:
       self._should_save_checkpoints = False
@@ -118,7 +118,7 @@ class Trainer(object):
       (slots, opt_params) = opt.tree_init(weights)
       return (OptState(weights, slots, opt_params), state)
 
-    if fastmath.backend_name() == 'jax':
+    if fastmath.is_backend(fastmath.Backend.JAX):
       # JIT parameter initialization to avoid memory fragmentation
       new_opt_state_and_model_state = (
           fastmath.jit(new_opt_state_and_model_state))
@@ -212,7 +212,7 @@ class Trainer(object):
 
   @property
   def learning_rate(self):
-    with fastmath.use_backend('numpy'):
+    with fastmath.use_backend(fastmath.Backend.NUMPY):
       return self._lr_schedule(self._step)
 
   def reset(self, output_dir, init_checkpoint=None):
@@ -325,8 +325,8 @@ class Trainer(object):
     if self._should_log_now():
       for name, value in stat.items():
         # TODO(afrozm): value is a scalar, but sometimes JAX is crashing here
-        # with a device put array error complaning that it should be an array.
-        # On  multiple devices, take the mean.
+        # with a device put array error complaining that it should be an array.
+        # On multiple devices, take the mean.
         scalar_value = np.mean(np.array(value))
         self._train_sw.scalar('training/' + name, scalar_value, step=self._step)
     self._step += 1
@@ -407,7 +407,7 @@ class Trainer(object):
       opt_state = OptState(*fastmath.nested_map(first_replica, opt_state))
     # This line, while optional, allows JAX to transfer arrays from the device
     # to the host in parallel, which is particularly important for cloud TPU.
-    if fastmath.backend_name() == 'jax':
+    if fastmath.is_backend(fastmath.Backend.JAX):
       opt_state = jax.device_get(opt_state)
     step, history, model_state = self._step, self._history, self._model_state
     output_dir = self._output_dir
@@ -489,7 +489,7 @@ class Trainer(object):
             and (self._step == 1 or self._step % 10 == 0))
 
   def _for_n_devices(self, x):
-    """Replicates/broadcasts `x` for n devices if `self.n_devicess > 1`."""
+    """Replicates/broadcasts `x` for n devices if `self.n_devices > 1`."""
     return tl.for_n_devices(x, self.n_devices)  # pylint: disable=protected-access
 
   def close(self):
@@ -518,7 +518,9 @@ def train(output_dir,
           metrics=None,
           checkpoint_highest=None,
           checkpoint_lowest=None,
-          use_loop=False):
+          use_loop=True,
+          loss_chunk_size=0,
+          use_memory_efficient_trainer=False):
   """Train the model on the inputs.
 
   Args:
@@ -544,6 +546,8 @@ def train(output_dir,
     checkpoint_highest: save the checkpoint highest at this metric.
     checkpoint_lowest: save the checkpoint lowest at this metric.
     use_loop: whether to use training.Loop instead of Trainer.
+    loss_chunk_size: int, if > 0 chunk loss into these sizes to save memory.
+    use_memory_efficient_trainer: whether to use memory-efficient trainer.
 
   Returns:
     trax.TrainerState or training.Loop if use_loop is True
@@ -555,9 +559,10 @@ def train(output_dir,
     # Inputs is either an Inputs instance or a function that returns it.
     if callable(inputs):  # If we pass a function, e.g., through gin, call it.
       inputs = inputs()
+    opt = optimizer if use_memory_efficient_trainer else optimizer()
     train_task = training.TrainTask(inputs.train_stream(n_devices),
                                     loss_layer=loss_fn,
-                                    optimizer=optimizer(),
+                                    optimizer=opt,
                                     lr_schedule=lr_schedule_fn(),
                                     n_steps_per_checkpoint=eval_frequency)
 
@@ -573,14 +578,17 @@ def train(output_dir,
     checkpoint_at = None
     if checkpoints_at is not None:
       checkpoint_at = lambda step: step in checkpoints_at
-    loop = training.Loop(model(mode='train'),
-                         [train_task],
-                         eval_model=model(mode='eval'),
-                         eval_tasks=[eval_task],
-                         output_dir=output_dir,
-                         checkpoint_at=checkpoint_at,
-                         n_devices=n_devices,
-                         random_seed=random_seed)
+    loop = training.Loop(
+        model(mode='train'),
+        [train_task],
+        eval_model=model(mode='eval'),
+        eval_tasks=[eval_task],
+        output_dir=output_dir,
+        checkpoint_at=checkpoint_at,
+        n_devices=n_devices,
+        loss_chunk_size=loss_chunk_size,
+        use_memory_efficient_trainer=use_memory_efficient_trainer,
+        random_seed=random_seed)
 
     # Train and return the loop.
     loop.run(steps)
@@ -611,7 +619,7 @@ def train(output_dir,
       # Bookkeeping we do at the first step
       if trainer.step == 1:
         # Save computation graph (single-device only for now)
-        if (save_graphs and fastmath.backend_name() == 'jax'):
+        if (save_graphs and fastmath.is_backend(fastmath.Backend.JAX)):
           trainer.save_computation_graphs()
 
         # Save Gin config
@@ -655,7 +663,7 @@ def _jit_update_fn(predict_fn, loss_fn, optimizer, n_devices, jit=True):
       return single_update
 
   # Else, for n_devices > 1:
-  @functools.partial(fastmath.pmap, axis_name='batch', donate_argnums=(0,))
+  @functools.partial(fastmath.pmap, axis_name='batch')  # donate_argnums=(0,))
   def mapped_update(weights_and_slots, i, opt_params, batch, state, rng):
     """This is a multi-device version of the update function above."""
     # We assume all tensors have the first dimension = n_devices.
