@@ -15,10 +15,9 @@
 
 """Tests for trax.supervised.trainer_lib."""
 
-import contextlib
 import functools
 import os
-import tempfile
+from absl.testing import absltest
 from absl.testing import parameterized
 
 from jax import test_util  # pylint: disable=unused-import
@@ -26,17 +25,17 @@ from jax.config import config
 from jax.lib import xla_bridge
 
 import tensorflow.compat.v2 as tf
-from tensorflow.compat.v2 import test
-from tensorflow.compat.v2.io import gfile
 
 from trax import fastmath
-from trax import layers
+from trax import layers as tl
 from trax import models
 from trax import optimizers as trax_opt
+from trax import test_utils
 from trax.data import inputs as inputs_lib
 from trax.fastmath import numpy as jnp
 from trax.supervised import lr_schedules as lr
 from trax.supervised import trainer_lib
+from trax.tf_numpy import extensions as npe
 from trax.tf_numpy import numpy as tf_np
 
 
@@ -67,111 +66,147 @@ def _test_inputs(n_classes, with_weights=False, input_shape=(6, 6, 3)):
   return inputs_lib.Inputs(input_stream_masked)
 
 
+def _test_inputs_lm(vocab_size, seq_len, per_device_batch_size=2):
+  """Make trainer_lib.inputs.Inputs for language model."""
+  batch_size = per_device_batch_size * xla_bridge.device_count()
 
-BACKENDS = ['jax', 'tf']
+  def input_stream(_):
+    def make_batch(key):
+      return fastmath.random.randint(
+          key, [batch_size, seq_len], dtype=jnp.int32, minval=0,
+          maxval=vocab_size)
+    key = fastmath.random.get_prng(0)
+    while True:
+      keys = fastmath.random.split(key, 3)
+      key = keys[0]
+      inputs = make_batch(keys[1])
+      targets = make_batch(keys[2])
+      yield inputs, targets
+
+  def input_stream_masked(n_devices):
+    return inputs_lib.add_loss_weights(input_stream(n_devices))
+
+  return inputs_lib.Inputs(input_stream_masked)
 
 
-class TraxTest(test.TestCase, parameterized.TestCase):
 
-  @contextlib.contextmanager
-  def tmp_dir(self):
-    tmp = tempfile.mkdtemp(dir=self.get_temp_dir())
-    yield tmp
-    gfile.rmtree(tmp)
+BACKENDS = [fastmath.Backend.JAX, fastmath.Backend.TFNP]
 
-  # TODO(wangpeng): Remove `skipTest`'s when tf-numpy's `pmap` is in place
 
-  def _test_train_eval_predict(self, backend_name):
-    if xla_bridge.device_count() > 1 and backend_name == 'tf':
-      self.skipTest("tf-numpy backend does't support multi-devices yet.")
-    with fastmath.use_backend(backend_name), self.tmp_dir() as output_dir:
+def short_name(b):
+  if b == fastmath.Backend.JAX:
+    return 'jax'
+  else:
+    return 'tf'
+
+
+def opt_name(opt):
+  if opt is None:
+    return 'None'
+  return opt.__name__
+
+
+class TraxTest(parameterized.TestCase):
+
+  def __init__(self, methodName='runTest'):  # pylint: disable=invalid-name
+    super().__init__(methodName)
+    if npe.tpu_devices():
+      # Initialize TPU for TF
+      resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu='local')
+      tf.tpu.experimental.initialize_tpu_system(resolver)
+
+  def setUp(self):
+    super().setUp()
+    test_utils.ensure_flag('test_tmpdir')
+    self._old_is_allow_float64 = tf_np.is_allow_float64()
+    tf_np.set_allow_float64(False)
+
+  def tearDown(self):
+    tf_np.set_allow_float64(self._old_is_allow_float64)
+    super().tearDown()
+
+  def _test_train_eval_predict(self, backend, model_name='Simple',
+                               optimizer=None):
+    with fastmath.use_backend(backend):
       # Prepare model and inputs
-      n_classes = 4
       steps = 2
       eval_steps = 2
 
-      # Adds Dropout and BatchNorm to test state handling.
-      def model_fn(mode='train'):
-        return layers.Serial(
-            layers.Dropout(mode=mode, rate=0.1), layers.BatchNorm(mode=mode),
-            models.MLP(d_hidden=16, n_output_classes=n_classes, mode=mode))
+      if model_name == 'Simple':
+        n_classes = 4
+        # Adds Dropout and BatchNorm to test state handling.
+        def model_fn(mode='train'):
+          return tl.Serial(
+              tl.Dropout(mode=mode, rate=0.1), tl.BatchNorm(mode=mode),
+              models.MLP(d_hidden=16, n_output_classes=n_classes, mode=mode))
+        inputs = _test_inputs(n_classes)
+        n_in = 1
+      elif model_name == 'Resnet50':
+        n_classes = 4
+        model_fn = models.Resnet50
+        inputs = _test_inputs(n_classes, input_shape=(224, 224, 3))
+        n_in = 1
+      elif model_name == 'Transformer':
+        vocab_size = 32
+        seq_len = 16
+        inputs = _test_inputs_lm(vocab_size, seq_len)
+        model_fn = functools.partial(
+            models.Transformer,
+            input_vocab_size=vocab_size)
+        n_in = 2
+      else:
+        raise ValueError('Unrecognized model name: ' + model_name)
 
-      inputs = _test_inputs(n_classes)
+      kwargs = {}
+      if optimizer is not None:
+        kwargs['optimizer'] = optimizer
 
       # Train and evaluate
-      state = trainer_lib.train(
-          output_dir,
-          model=model_fn,
-          inputs=inputs,
-          steps=steps,
-          eval_steps=eval_steps)
-
-      # Assert total train steps
-      self.assertEqual(steps, state.step)
-
-      # Assert 2 evaluations ran
-      train_acc = state.history.get('train', 'metrics/accuracy')
-      eval_acc = state.history.get('eval', 'metrics/accuracy')
-      self.assertEqual(len(train_acc), len(eval_acc))
-      self.assertLen(eval_acc, 2)
-
-      # Predict with final weights
-      inputs = inputs.train_stream(1)
-      model = model_fn()
-      weights = state.opt_state.weights[0]
-      state = state.model_state[0]
-      if xla_bridge.device_count() > 1:
-        unreplicate = lambda x: x[0]
-        weights = fastmath.nested_map(unreplicate, weights)
-        state = fastmath.nested_map(unreplicate, state)
-      model(next(inputs)[0], weights=weights, state=state)
-
-  @parameterized.parameters(BACKENDS)
-  def test_train_eval_predict(self, backend_name):
-    self._test_train_eval_predict(backend_name)
-
-  @parameterized.parameters(BACKENDS)
-  def test_train_eval_predict_sm3(self, backend_name):
-    if xla_bridge.device_count() > 1 and backend_name == 'tf':
-      self.skipTest("tf-numpy backend doesn't support multi-devices yet.")
-    with fastmath.use_backend(backend_name), self.tmp_dir() as output_dir:
-      # Prepare model and inputs
-      n_classes = 4
-      steps = 2
-      eval_steps = 2
-      model_fn = functools.partial(
-          models.MLP, d_hidden=16, n_output_classes=n_classes)
-      inputs = _test_inputs(n_classes)
-
-      # Train and evaluate
-      state = trainer_lib.train(
+      output_dir = self.create_tempdir().full_path
+      loop = trainer_lib.train(
           output_dir,
           model=model_fn,
           inputs=inputs,
           steps=steps,
           eval_steps=eval_steps,
-          optimizer=trax_opt.SM3)
+          eval_frequency=1,  # eval at every step.
+          **kwargs)
 
       # Assert total train steps
-      self.assertEqual(steps, state.step)
+      self.assertEqual(steps, loop.step)
 
-      # Assert 2 evaluations ran
-      train_acc = state.history.get('train', 'metrics/accuracy')
-      eval_acc = state.history.get('eval', 'metrics/accuracy')
-      self.assertEqual(len(train_acc), len(eval_acc))
-      self.assertLen(eval_acc, 2)
+      inputs = inputs.train_stream(1)
+
+      # Predict with final weights
+      model = model_fn()
+      weights = loop.model.weights
+      state = loop.model.state
+      model(next(inputs)[:n_in], weights=weights, state=state)
 
       # Predict with weights loaded from file.
-      inputs = inputs.train_stream(1)
       model = model_fn()
       model.init_from_file(os.path.join(output_dir, 'model.pkl.gz'))
-      model(next(inputs)[0])
+      model(next(inputs)[:n_in])
+
+  @parameterized.named_parameters(
+      ('_%s_%s_%s' % (short_name(backend), model_name, opt_name(opt)),  # pylint: disable=g-complex-comprehension
+       backend, model_name, opt)
+      for backend, configs in [
+          (fastmath.Backend.JAX, [('Simple', None)]),
+          (fastmath.Backend.TFNP, [('Simple', None),
+                                   ('Resnet50', trax_opt.Momentum),
+                                   ('Transformer', trax_opt.Adam)])]
+      for model_name, opt in configs)
+  def test_train_eval_predict(self, backend, model_name, opt):
+    self._test_train_eval_predict(backend, model_name, opt)
 
   @parameterized.parameters(BACKENDS)
-  def test_train_restart(self, backend_name):
-    if xla_bridge.device_count() > 1 and backend_name == 'tf':
-      self.skipTest("tf-numpy backend doesn't support multi-devices yet.")
-    with fastmath.use_backend(backend_name), self.tmp_dir() as output_dir:
+  def test_train_eval_predict_sm3(self, backend):
+    self._test_train_eval_predict(backend, 'Simple', trax_opt.SM3)
+
+  @parameterized.parameters(BACKENDS)
+  def test_train_restart(self, backend):
+    with fastmath.use_backend(backend):
       # Prepare model and inputs
       n_classes = 4
       steps = 2
@@ -181,29 +216,31 @@ class TraxTest(test.TestCase, parameterized.TestCase):
       inputs = _test_inputs(n_classes)
 
       # Train and evaluate
+      output_dir = self.create_tempdir().full_path
       trainer_lib.train(
           output_dir,
           model=model_fn,
           inputs=inputs,
           steps=steps,
-          eval_steps=eval_steps)
+          eval_steps=eval_steps,
+          eval_frequency=1)
 
       # Restart training
-      state = trainer_lib.train(
+      loop = trainer_lib.train(
           output_dir,
           model=model_fn,
           inputs=inputs,
           steps=(2 * steps),
-          eval_steps=eval_steps)
+          eval_steps=eval_steps,
+          eval_frequency=1)
 
-      # Assert total train steps
-      self.assertEqual(state.step, 2 * steps)
+      # Assert total train steps - with loop we don't resume, but train for as
+      # many steps as given, so: steps + 2*steps = 3*steps.
+      self.assertEqual(loop.step, 3 * steps)
 
   @parameterized.parameters(BACKENDS)
-  def test_train_with_weights(self, backend_name):
-    if xla_bridge.device_count() > 1 and backend_name == 'tf':
-      self.skipTest("tf-numpy backend doesn't support multi-devices yet.")
-    with fastmath.use_backend(backend_name), self.tmp_dir() as output_dir:
+  def test_train_with_weights(self, backend):
+    with fastmath.use_backend(backend):
       # Prepare model and inputs
       n_classes = 4
       steps = 2
@@ -213,6 +250,7 @@ class TraxTest(test.TestCase, parameterized.TestCase):
       inputs = _test_inputs(n_classes, with_weights=True)
 
       # Train and evaluate
+      output_dir = self.create_tempdir().full_path
       state = trainer_lib.train(
           output_dir,
           model=model_fn,
@@ -224,11 +262,8 @@ class TraxTest(test.TestCase, parameterized.TestCase):
       self.assertEqual(state.step, steps)
 
   @parameterized.parameters(BACKENDS)
-  def test_reset_twice(self, backend_name):
-    if xla_bridge.device_count() > 1 and backend_name == 'tf':
-      self.skipTest("tf-numpy backend doesn't support multi-devices yet.")
-    with fastmath.use_backend(backend_name), self.tmp_dir() as output_dir1, \
-          self.tmp_dir() as output_dir2:
+  def test_reset_twice(self, backend):
+    with fastmath.use_backend(backend):
       n_classes = 4
       model_fn = functools.partial(
           models.MLP, d_hidden=16, n_output_classes=n_classes)
@@ -236,14 +271,16 @@ class TraxTest(test.TestCase, parameterized.TestCase):
 
       trainer = trainer_lib.Trainer(
           model=model_fn,
-          loss_fn=layers.CrossEntropyLoss(),
+          loss_fn=tl.CrossEntropyLoss(),
           optimizer=trax_opt.SM3,
           lr_schedule=lr.multifactor(),
           inputs=inputs,
       )
 
+      output_dir1 = self.create_tempdir(name='output_dir1').full_path
       trainer.reset(output_dir1)
       trainer.evaluate(1)
+      output_dir2 = self.create_tempdir(name='output_dir2').full_path
       trainer.reset(output_dir2)
       trainer.evaluate(1)
 
@@ -262,20 +299,19 @@ class TraxTest(test.TestCase, parameterized.TestCase):
     computation to copy int32/uint32 outputs to CPU. This test makes sure that
     won't happen.
     """
-    if xla_bridge.device_count() > 1:
-      self.skipTest("tf-numpy backend doesn't support multi-devices yet.")
-    with fastmath.use_backend('tf'), self.tmp_dir() as output_dir:
+    with fastmath.use_backend(fastmath.Backend.TFNP):
       n_classes = 1001
       model_fn = functools.partial(models.Resnet50,
                                    n_output_classes=n_classes)
       inputs = _test_inputs(n_classes, input_shape=(224, 224, 3))
       trainer = trainer_lib.Trainer(
           model=model_fn,
-          loss_fn=layers.CrossEntropyLoss(),
+          loss_fn=tl.CrossEntropyLoss(),
           optimizer=trax_opt.SM3,
           lr_schedule=lr.multifactor(),
           inputs=inputs,
       )
+      output_dir = self.create_tempdir().full_path
       trainer.reset(output_dir)
       trainer.train_epoch(1, 0)
       # Those are the things returned by Trainer._jit_update_fn
@@ -289,7 +325,7 @@ class TraxTest(test.TestCase, parameterized.TestCase):
 
 
 
-class EpochsTest(test.TestCase):
+class EpochsTest(absltest.TestCase):
 
   def test_cuts_epoch_when_total_steps_reached(self):
     epoch_steps = trainer_lib.epochs(
@@ -310,4 +346,4 @@ class EpochsTest(test.TestCase):
 if __name__ == '__main__':
   config.config_with_absl()
   tf.compat.v1.enable_eager_execution()
-  test.main()
+  absltest.main()
