@@ -21,6 +21,10 @@ from __future__ import print_function
 
 import bisect
 import contextlib
+import copy
+import functools
+import string
+import sys
 import threading
 import numpy as np
 import six
@@ -36,6 +40,9 @@ _int_dtypes = [
     tf.int64, tf.int32, tf.int16, tf.int8, tf.uint8, tf.uint16, tf.uint32,
     tf.uint64
 ]
+_tf_nn_APIs = {1: [tf.nn.conv1d, tf.nn.conv1d_transpose],
+               2: [tf.nn.conv2d, tf.nn.conv2d_transpose],
+               3: [tf.nn.conv3d, tf.nn.conv3d_transpose]}
 
 
 def most_precise_int_dtype(x):
@@ -264,16 +271,20 @@ def grad(f, has_aux=False):
   return _f
 
 
-# A workaround for b/121383831
-_orig_result_is_list = threading.local()
+def _record_result_type(recorder, f):
+  """A decorator that records some information about the function.
 
+  Args:
+    recorder: a function of signature `(args, kwargs, res) -> res`.
+    f: the original function.
 
-def _record_result_type(f):
-  # A wrapper just for setting _orig_result_is_list, as a workaround for
-  # b/121383831
+  Returns:
+    A transformed function that calls the original function and then the
+    recorder afterwards.
+  """
   def wrapper(*args, **kwargs):
     res = f(*args, **kwargs)
-    _orig_result_is_list.val = isinstance(res, list)
+    res = recorder(args, kwargs, res)
     return res
 
   return wrapper
@@ -283,7 +294,8 @@ def jit(f,
         static_argnums=(),
         xla_forced_compile=False,
         input_signature=None,
-        autograph=False):
+        autograph=False,
+        experimental_compile=False):
   """Returns a function that runs a trace-compiled version of `f`.
 
   A trace-compiled version of a function `f` has the same behavior as `f` (when
@@ -310,7 +322,8 @@ def jit(f,
       static arguments. Note that as aforementioned, any arguments that were not
       convertible to tensor will also be static.
     xla_forced_compile: if true, it will use XLA to force-compile the graph.
-      This requires that the function only contain ops that are XLA compatible.
+      This requires that the function only contain ops that are XLA
+      compatible. It will compile the entire function into a single XLA op.
     input_signature: a list of `tf.TensorSpec`, as the input signature to
       control tracing behavior. See the
       [doc](https://www.tensorflow.org/api_docs/python/tf/function]) of
@@ -319,23 +332,49 @@ def jit(f,
       `if` and `while` to their TensorFlow counterparts. See the
       [doc](https://www.tensorflow.org/api_docs/python/tf/function]) of
         `tf.function` for details.
+    experimental_compile: the `experimental_compile` flag for `tf.function`. See
+      the [doc](https://www.tensorflow.org/api_docs/python/tf/function]) of
+      `tf.function` for details. This is the recommended way to turn on XLA for
+      tf.function, but unlike xla_forced_compile, it doesn't force-compile the
+      entire function into a single XLA op.
 
   Returns:
     A trace-compiled version of f.
   """
 
-  @tf.function(input_signature=input_signature, autograph=autograph)
+  @tf.function(input_signature=input_signature, autograph=autograph,
+               experimental_compile=experimental_compile)
   def _tf_f(*args, **kwargs):
     """Accelerated function with tensor inputs/outputs."""
     np_args = _tf_to_np(args)
     kwargs = {k: _tf_to_np(v) for k, v in kwargs.items()}
     if xla_forced_compile:
-      # Workaround b/121383831
-      f_ = _record_result_type(f)
+      # Use list for mutability
+      output_is_list = [False]
+      output_is_empty = [False]
+      output_structure = [None]
+      def recorder(args, kwargs, res):
+        del args, kwargs
+        # Workaround b/121383831
+        output_is_list[0] = isinstance(res, list)
+        # If outputs are empty, xla.compile returns an `Operation`, which we
+        # don't want.
+        if tf.nest.flatten(res):
+          output_is_empty[0] = False
+          output_structure[0] = None
+        else:
+          output_is_empty[0] = True
+          # Without deepcopy, xla.compile will change output_structure[0] to a
+          # list of `Operation`.
+          output_structure[0] = copy.deepcopy(res)
+        return res
+      f_ = _record_result_type(recorder, f)
       np_out = tf.xla.experimental.compile(lambda: f_(*np_args, **kwargs))
       # Workaround b/121383831
-      if (isinstance(np_out, list) and len(np_out) == 1 and
-          not _orig_result_is_list.val):
+      if output_is_empty[0]:
+        np_out = output_structure[0]
+      elif (isinstance(np_out, list) and len(np_out) == 1 and
+            not output_is_list[0]):
         np_out = np_out[0]
     else:
       np_out = f(*np_args, **kwargs)
@@ -355,7 +394,7 @@ def jit(f,
   return _f
 
 
-def eval_on_shapes(f, static_argnums=()):
+def eval_on_shapes(f, static_argnums=(), allow_static_outputs=False):
   """Returns a function that evaluates `f` given input shapes and dtypes.
 
   It transforms function `f` to a function that performs the same computation as
@@ -363,13 +402,62 @@ def eval_on_shapes(f, static_argnums=()):
 
   Args:
     f: the function to be transformed.
-    static_argnums: See documentation of `jit`.
+    static_argnums: see documentation of `jit`.
+    allow_static_outputs: whether to allow non-array outputs. If True, non-array
+      outputs (e.g. Python integers) will be returned as-is; otherwise, they
+      will be converted to ndarrays, and then specs of those ndarrays will be
+      returned.
 
   Returns:
     A function whose input arguments can be either the same as `f`'s or only
     their shapes/dtypes represented by `tf.TensorSpec`, and whose return values
-    are `tf.TensorSpec`s with the same nested structure as `f`'s return values.
+    are `tf.TensorSpec`s with the same nested structure as `f`'s return
+    values. If `allow_static_outputs` is True, when `f` returns some non-array
+    outputs (e.g. Python integers), the converted function will return them
+    as-is instead of returning `tf.TensorSpec`s for them.
   """
+  def abstractify(args):
+    def _abstractify(x):
+      x = _canonicalize_jit_arg(x)
+      if isinstance(x, (tf.Tensor, tf_np.ndarray)):
+        return tf.TensorSpec(x.shape, x.dtype)
+      else:
+        return x
+    new_args = []
+    for i, arg in enumerate(args):
+      if i in static_argnums:
+        new_args.append(arg)
+      else:
+        new_args.append(tf.nest.map_structure(_abstractify, arg))
+    return new_args
+
+  if allow_static_outputs:
+    # When `tf_f` below is called (via get_concrete_function) with the same
+    # arugments (after abstraction), the Python function `f` won't be run, so we
+    # need this python_outputs_map to retrieve the Python outputs we've seen
+    # before that correspond the arguments.
+    python_outputs_map = {}
+    def recorder(args, kwargs, res):
+      # Since the get_concrete_function below only uses positional args, we also
+      # only positional args here.
+      del args, kwargs
+      def is_tensor_like(x):
+        if hasattr(x, "_type_spec"):
+          return True  # x is a CompositeTensor
+        return isinstance(x, (tf_np.ndarray, tf.Tensor))
+      py_values = tf.nest.map_structure(
+          lambda x: None if is_tensor_like(x) else x,
+          res)
+      key = id(tf.compat.v1.get_default_graph())
+      python_outputs_map[key] = py_values
+      # Set non-tensor outputs to None to avoid tf.function calling
+      # tf.convert_to_tensor on them.
+      res = tf.nest.map_structure(
+          lambda x: None if not is_tensor_like(x) else x,
+          res)
+      return res
+    f = _record_result_type(recorder, f)
+
   # TODO(wangpeng): tf.function could add a knob to turn off materializing the
   #   graph, so that we don't waste computation and memory when we just want
   #   shape inference.
@@ -377,35 +465,127 @@ def eval_on_shapes(f, static_argnums=()):
 
   # pylint: disable=missing-docstring
   def f_return(*args):
-
-    def abstractify(x):
-      x = _canonicalize_jit_arg(x)
-      if isinstance(x, (tf.Tensor, tf_np.ndarray)):
-        return tf.TensorSpec(x.shape, x.dtype)
-      else:
-        return x
-
     def to_tensor_spec(x):
       if isinstance(x, tf.Tensor):
         return tf.TensorSpec(x.shape, x.dtype)
       else:
         return x
 
-    new_args = []
-    for i, arg in enumerate(args):
-      if i in static_argnums:
-        new_args.append(arg)
-      else:
-        new_args.append(tf.nest.map_structure(abstractify, arg))
-    res = tf_f.get_concrete_function(*new_args).structured_outputs
+    new_args = abstractify(args)
+    cfun = tf_f.get_concrete_function(*new_args)
+    res = cfun.structured_outputs
+    res = tf.nest.map_structure(to_tensor_spec, res)
 
-    return tf.nest.map_structure(to_tensor_spec, res)
+    if allow_static_outputs:
+      key = id(cfun.graph)
+      py_values = python_outputs_map[key]
+      # We can also call tf.get_static_value on structured_outputs to retrieve
+      # the Python values, but since we'll need to use python_outputs_map to
+      # record "which outputs are static?" anyway, we choose to directly store
+      # the Python values in python_outputs_map.
+      res = tf.nest.map_structure(
+          lambda x, python_value: x if python_value is None else python_value,
+          res, py_values)
+
+    return res
 
   # Provides access to `tf_f` for testing purpose.
   f_return._tf_function = tf_f  # pylint: disable=protected-access
   return f_return
 
 
+def _index_update_helper(updater, x, idx, y):
+  x = tf_np.asarray(x)
+  y = tf_np.asarray(y)
+  # TODO(b/164251540): Remove this expensive manual broadcasting once
+  #   tf.raw_ops.tensor_strided_slice_update and tf.tensor_scatter_nd_update
+  #   support broadcasting.
+  y = tf.broadcast_to(y.data, tf.shape(x[idx].data))
+  return updater(x, idx, y)
+
+
+# pylint: disable=protected-access
+def index_update(x, idx, y):
+  """Pure equivalent of `x[idx] = y`.
+
+  Returns the value of x that would result from the NumPy-style indexed
+  assignment `x[idx] = y`. Because it's a pure function, `x` itself won't be
+  changed.
+
+  Args:
+    x: an array with the values to be updated.
+    idx: a Numpy-style index, consisting of `None`, integers, slice objects,
+      ellipses, ndarrays with integer dtypes, or a tuple of the above.
+    y: the array of updates. `y` must be broadcastable to the shape of the array
+      that would be returned by `x[idx]`.
+
+  Returns:
+    The updated version of `x`.
+  """
+  return _index_update_helper(tf_np.ndarray._with_index_update, x, idx, y)
+
+
+def index_add(x, idx, y):
+  """Pure equivalent of `x[idx] += y`.
+
+  Returns the value of x that would result from the NumPy-style indexed
+  assignment `x[idx] += y`. Because it's a pure function, `x` itself won't be
+  changed.
+
+  Args:
+    x: an array with the values to be updated.
+    idx: a Numpy-style index, consisting of `None`, integers, slice objects,
+      ellipses, ndarrays with integer dtypes, or a tuple of the above.
+    y: the array of updates. `y` must be broadcastable to the shape of the array
+      that would be returned by `x[idx]`.
+
+  Returns:
+    The updated version of `x`.
+  """
+  return _index_update_helper(tf_np.ndarray._with_index_add, x, idx, y)
+
+
+def index_min(x, idx, y):
+  """Pure equivalent of `x[idx] = minimum(x[idx], y)`.
+
+  Returns the value of x that would result from the NumPy-style indexed
+  assignment `x[idx] = minimum(x[idx], y)`. Because it's a pure function, `x`
+  itself won't be changed.
+
+  Args:
+    x: an array with the values to be updated.
+    idx: a Numpy-style index, consisting of `None`, integers, slice objects,
+      ellipses, ndarrays with integer dtypes, or a tuple of the above.
+    y: the array of updates. `y` must be broadcastable to the shape of the array
+      that would be returned by `x[idx]`.
+
+  Returns:
+    The updated version of `x`.
+  """
+  return _index_update_helper(tf_np.ndarray._with_index_min, x, idx, y)
+
+
+def index_max(x, idx, y):
+  """Pure equivalent of `x[idx] = maximum(x[idx], y)`.
+
+  Returns the value of x that would result from the NumPy-style indexed
+  assignment `x[idx] = maximum(x[idx], y)`. Because it's a pure function, `x`
+  itself won't be changed.
+
+  Args:
+    x: an array with the values to be updated.
+    idx: a Numpy-style index, consisting of `None`, integers, slice objects,
+      ellipses, ndarrays with integer dtypes, or a tuple of the above.
+    y: the array of updates. `y` must be broadcastable to the shape of the array
+      that would be returned by `x[idx]`.
+
+  Returns:
+    The updated version of `x`.
+  """
+  return _index_update_helper(tf_np.ndarray._with_index_max, x, idx, y)
+# pylint: enable=protected-access
+
+
 def logsumexp(x, axis=None, keepdims=None):
   """Computes log(sum(exp(elements across dimensions of a tensor))).
 
@@ -443,6 +623,248 @@ def erf(x):
   return tf_np.asarray(tf.math.erf(x.data))
 
 
+def _minus(a, b):
+  return [x for x in a if x not in b]
+
+
+def _compose_output_rep(lhs_rep, rhs_rep, lhs_contraction, rhs_contraction,
+                        lhs_batch, rhs_batch):
+  """Compose the output string representation.
+
+  e.g., ij, jk, (((1,), (0,)), ((), ())) -> ik
+        aij, ajk, (((2,), (1,)), ((0,), (0,))) -> aik
+
+  Args:
+    lhs_rep: A string representation for the left-hand side input array
+    rhs_rep: A string representation for the right-hand side input array
+    lhs_contraction: Sequence[int] (the contraction dimensions of lhs)
+    rhs_contraction: Sequence[int] (the contraction dimensions of rhs)
+    lhs_batch: Sequence[int] (the batch dimensions of lhs)
+    rhs_batch: Sequence[int] (the batch dimensions of rhs)
+
+  Returns:
+    A string representation of the result array.
+  """
+  output_rep = []
+  for dim in lhs_batch:
+    output_rep.append(lhs_rep[dim])
+
+  for i in _minus(range(len(lhs_rep)), lhs_batch + lhs_contraction):
+    output_rep.append(lhs_rep[i])
+  for i in _minus(range(len(rhs_rep)), rhs_batch + rhs_contraction):
+    output_rep.append(rhs_rep[i])
+  return "".join(output_rep)
+
+
+def _non_batched_matmul(lhs, rhs, lhs_contraction, rhs_contraction):
+  """Compute the non-batched matrix multiplication.
+
+  If it is the general non-batched/single-batched matrix multiplication,
+  use the highly optimized kernel `tf.tensordot` to handle it.
+
+  Args:
+    lhs: an array (the left-hand side matrix/vector to be multiplied)
+    rhs: an array (the right-hand side matrix/vector to be multiplied)
+    lhs_contraction: Sequence[int] (the contraction dimensions of lhs)
+    rhs_contraction: Sequence[int] (the contraction dimensions of rhs)
+
+  Returns:
+    An array that contains the result.
+  """
+  return tf.tensordot(
+      lhs, rhs, axes=(list(lhs_contraction), list(rhs_contraction)))
+
+
+def tf_dot_general(lhs, rhs, dimension_numbers):
+  """The general dot operation for TensorFlow.
+
+  An equivalent general dot operation as that in JAX -
+     <https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.dot_general.html>
+  Although there is an implementation in TF XLA, avoid directly using XLA when
+  possible.
+
+  e.g., non-batched: ij,jk->ik
+        batched: ijk,ikl->ijl
+
+  Args:
+    lhs: an array (the left-hand side matrix/vector to be multiplied)
+    rhs: an array (the right-hand side matrix/vector to be multiplied)
+    dimension_numbers: (Tuple[Tuple[Sequence[int], Sequence[int]],
+      Tuple[Sequence[int], Sequence[int]]]) – a tuple of tuples of the form
+      ((lhs_contracting_dims, rhs_contracting_dims), (lhs_batch_dims,
+      rhs_batch_dims))
+
+  Returns:
+    An array that contains the result.
+  """
+  char_list = list(string.ascii_lowercase)
+  char_list = char_list[8:] + char_list[:8]
+  lhs_rank, rhs_rank = len(lhs.shape), len(rhs.shape)
+  lhs_rep = char_list[:lhs_rank]
+  rhs_rep = char_list[lhs_rank:lhs_rank + rhs_rank]
+  contraction, batch = dimension_numbers
+  lhs_contraction, rhs_contraction = contraction
+  if len(lhs_contraction) != len(rhs_contraction):
+    raise ValueError(
+        "The input matrices are required to have the same number "
+        "of contraction dimensions, but got: lhs {}, rhs: {}".format(
+            len(lhs_contraction), len(rhs_contraction)))
+  lhs_batch, rhs_batch = batch
+  if len(lhs_batch) != len(rhs_batch):
+    raise ValueError("The input matrices are required to have the same number "
+                     "of batch dimensions, but got: lhs {}, rhs: {}".format(
+                         len(lhs_batch), len(rhs_batch)))
+
+  if not lhs_batch and not rhs_batch:
+    return _non_batched_matmul(lhs, rhs, lhs_contraction, rhs_contraction)
+
+  if (lhs_rank == rhs_rank == 3 and lhs_batch == (0,) and rhs_batch == (0,) and
+      lhs_contraction == (2,) and rhs_contraction == (1,)):
+    return tf.linalg.matmul(lhs, rhs)
+
+  for i in range(len(lhs_contraction)):
+    rhs_rep[rhs_contraction[i]] = lhs_rep[lhs_contraction[i]]
+  for i in range(len(lhs_batch)):
+    rhs_rep[rhs_batch[i]] = lhs_rep[lhs_batch[i]]
+
+  output_rep = _compose_output_rep(lhs_rep, rhs_rep, lhs_contraction,
+                                   rhs_contraction, lhs_batch, rhs_batch)
+  equation = "".join(lhs_rep) + "," + "".join(rhs_rep) + "->" + output_rep
+  return tf.einsum(equation, lhs, rhs)
+
+
+def _conv_general_param_type_converter(window_strides, lhs_dilation,
+                                       rhs_dilation, dim):
+  """Convert strides, lhs_dilation, rhs_dilation to match TF convention.
+
+  For example,
+   in the 3D case, if lhs_dilation = 2, then convert it to [2, 2, 2]
+                   if lhs_dilation = (2, 2, 2), convert it also to [2, 2, 2]
+
+  Args:
+    window_strides: window_strides to be converted
+    lhs_dilation: lhs_dilation to be converted
+    rhs_dilation: rhs_dilation to be converted
+    dim: dim to be converted
+
+  Returns:
+    The updated window_strides, lhs_dilation and rhs_dilation
+  """
+  def _as_list_of_size(item, size):
+    if item is None:
+      return None
+    return [item] * size if isinstance(item, int) else list(item)
+  return (_as_list_of_size(window_strides, dim),
+          _as_list_of_size(lhs_dilation, dim),
+          _as_list_of_size(rhs_dilation, dim))
+
+
+# pylint: disable=g-bad-todo
+# TODO(DarrenZhang01): Expand the test cases of general convolution and revise
+# the according bugs.
+# TODO(DarrenZhang01): Support feature_group_count, batch_group_count and
+# precision, and allow lhs_dilation and rhs_dilation to happen at the same time.
+# pylint: enable=g-bad-todo
+def tf_conv_general_dilated(lhs, rhs, window_strides, padding, output_shape,
+                            lhs_dilation=None, rhs_dilation=None,
+                            dimension_numbers=None, feature_group_count=1,
+                            batch_group_count=1, precision=None):
+  """A general conv API for TensorFlow.
+
+  According JAX version:
+    https://jax.readthedocs.io/en/stable/_autosummary/jax.lax.conv_general_dilated.html
+
+  Args:
+    lhs: a rank n+2 dimensional input array.
+    rhs: a rank n+2 dimensional array of kernel weights.
+    window_strides: a sequence of n integers, representing the inter-window
+                    strides.
+    padding: either the string ‘SAME’, the string ‘VALID’, or a sequence of n
+             (low, high) integer pairs that give the padding to apply before and
+             after each spatial dimension.
+    output_shape: the output shape of the convolution (only required for
+                  transpose convolution).
+    lhs_dilation: None, or a sequence of n integers, giving the dilation factor
+                  to apply in each spatial dimension of lhs. LHS dilation is
+                  also known as transposed convolution.
+    rhs_dilation: None, or a sequence of n integers, giving the dilation factor
+                  to apply in each spatial dimension of rhs. RHS dilation is
+                  also known as atrous convolution.
+    dimension_numbers: either None, a ConvDimensionNumbers object, or a 3-tuple
+                       (lhs_spec, rhs_spec, out_spec), where each element is a
+                       string of length n+2.
+    feature_group_count:  integer, default 1. Changing this is currently not
+                          supported.
+    batch_group_count: integer, default 1. Changing this is currently not
+                       supported.
+    precision: Optional. Either None, which means the default precision for the
+               backend, or a Precision enum value.
+
+  Returns:
+    A TF NumPy array that contains the convolution result.
+  """
+  dim = None
+  lhs_spec, rhs_spec, out_spec = dimension_numbers
+  if lhs_spec != out_spec:
+    raise ValueError("Current implementation requires the `data_format` of the "
+                     "inputs and outputs to be the same.")
+  if len(lhs_spec) >= 6:
+    raise ValueError("Current implmentation does not support 4 or higher"
+                     "dimensional convolution, but got: ", len(lhs_spec) - 2)
+  dim = len(lhs_spec) - 2
+  if lhs_dilation and rhs_dilation:
+    if lhs_dilation == (1,) * dim and rhs_dilation == (1,) * dim:
+      lhs_dilation, rhs_dilation = None, None
+    else:
+      raise ValueError("Current implementation does not support that "
+                       "deconvolution and dilation to be performed at the same "
+                       "time, but got lhs_dilation: {}, rhs_dilation: {}"
+                       .format(lhs_dilation, rhs_dilation))
+  if padding not in ["SAME", "VALID"]:
+    raise ValueError("Current implementation requires the padding parameter"
+                     "to be either 'VALID' or 'SAME', but got: ", padding)
+  if batch_group_count != 1 or feature_group_count != 1:
+    raise NotImplementedError("batch_group_count and feature_group_count "
+                              "other than 1 is currently not supported, but"
+                              " got feature_group_count: {}, batch_group_count"
+                              ": {}".format(feature_group_count,
+                                            batch_group_count))
+  if precision is not None:
+    raise NotImplementedError("precision other than `None` is currently not "
+                              "supported, but got: {}".format(precision))
+  # Convert params from int/Sequence[int] to list of ints.
+  strides, lhs_dilation, rhs_dilation = _conv_general_param_type_converter(
+      window_strides, lhs_dilation, rhs_dilation, dim
+  )
+  # Preprocess the shapes
+  dim_maps = {}
+  if isinstance(lhs_spec, str):
+    dim_maps["I"] = list(rhs_spec).index("I")
+    dim_maps["O"] = list(rhs_spec).index("O")
+    dim_maps["N"] = list(lhs_spec).index("N")
+    dim_maps["C"] = list(lhs_spec).index("C")
+  else:
+    dim_maps["I"] = rhs_spec[1]
+    dim_maps["O"] = rhs_spec[0]
+    dim_maps["N"] = lhs_spec[0]
+    dim_maps["C"] = lhs_spec[1]
+
+  lhs = tf_np.moveaxis(lhs, (dim_maps["N"], dim_maps["C"]), (0, dim + 1))
+  # Adjust the filters, put the dimension 'I' and 'O' at last.
+  rhs = tf_np.moveaxis(rhs, (dim_maps["O"], dim_maps["I"]), (dim + 1, dim))
+  spatial_dim_maps = {1: "W", 2: "HW", 3: "DHW"}
+  data_format = "N" + spatial_dim_maps[dim] + "C"
+
+  if rhs_dilation or (lhs_dilation is None and rhs_dilation is None):
+    output = _tf_nn_APIs[dim][0](lhs, rhs, strides, padding, data_format,
+                                 rhs_dilation)
+  else:
+    output = _tf_nn_APIs[dim][1](lhs, rhs, tf.constant(output_shape), strides,
+                                 padding, data_format, lhs_dilation)
+  output = tf_np.moveaxis(output, (0, dim + 1), (dim_maps["N"], dim_maps["C"]))
+  return output
+
+
 def conv(inp,
          fltr,
          window_strides,
@@ -484,8 +906,9 @@ def conv(inp,
                      "data format (%s)" % (input_spec, filter_spec))
   # No type promotion in order to prevent accidentally doing more expensive
   # computation.
-  inp = tf_np.asarray(inp)
-  fltr = tf_np.asarray(fltr)
+  dtype = tf_np.result_type(inp, fltr)
+  inp = tf_np.asarray(inp, dtype)
+  fltr = tf_np.asarray(fltr, dtype)
   return tf_np.asarray(
       tf.nn.convolution(
           input=inp.data,
@@ -616,7 +1039,8 @@ def scan(f, init, xs, length=None, reverse=False):
       that ``f`` accepts two arguments where the first is a value of the loop
       carry and the second is a slice of ``xs`` along its leading axis, and that
       ``f`` returns a pair where the first element represents a new value for
-      the loop carry and the second represents a slice of the output.
+      the loop carry and the second represents a slice of the output. Note that
+      the input and output carry must have the same dtype.
     init: an initial loop carry value of type ``c``, which can be a scalar,
       array, or any pytree (nested Python tuple/list/dict) thereof, representing
       the initial loop carry value. This value must have the same structure as
@@ -636,9 +1060,12 @@ def scan(f, init, xs, length=None, reverse=False):
     loop carry value and the second element represents the stacked outputs of
     the second output of ``f`` when scanned over the leading axis of the inputs.
   """
-  init, xs = tf.nest.map_structure(tf_np.asarray, (init, xs))
+  init, xs = tf.nest.map_structure(
+      lambda x: tf_np.asarray(x) if x is not None else None, (init, xs))
   init, xs = _np_to_tf((init, xs))
   def get_length(x):
+    if x is None:
+      return None
     if x.shape.rank == 0:
       raise ValueError("Some array in `xs` doesn't have a leading dimension")
     return x.shape[0]
@@ -654,25 +1081,53 @@ def scan(f, init, xs, length=None, reverse=False):
     raise ValueError(
         "Can't determine length. Please set the `length` argument.")
   xs_ta = tf.nest.map_structure(
-      lambda t: tf.TensorArray(t.dtype, size=0, dynamic_size=True).unstack(t),
+      lambda t: (tf.TensorArray(t.dtype, size=0, dynamic_size=True).unstack(t)  # pylint: disable=g-long-lambda
+                 if t is not None else None),
       xs)
-  def body(i, carry, ys_ta):
+  # tf.while_loop doesn't allow None in loop_vars, so we mask them.
+  is_init_none = tf.nest.map_structure(lambda x: x is None, init)
+  def to_safe(carry):
+    return tf.nest.map_structure(
+        lambda x, is_none: tf.zeros([]) if is_none else x, carry, is_init_none)
+  def from_safe(safe_carry):
+    return tf.nest.map_structure(
+        lambda x, is_none: None if is_none else x, safe_carry, is_init_none)
+  def body(i, safe_carry, ys_ta):
+    carry = from_safe(safe_carry)
     if reverse:
       i_ = length - 1 - i
     else:
       i_ = i
-    xs = tf.nest.map_structure(lambda x_ta: x_ta.read(i_), xs_ta)
+    xs = tf.nest.map_structure(
+        lambda x_ta: x_ta.read(i_) if x_ta is not None else None, xs_ta)
     carry, ys = _np_to_tf(f(*_tf_to_np((carry, xs))))
-    ys_ta = tf.nest.map_structure(lambda y_ta, y: y_ta.write(i_, y), ys_ta, ys)
-    return i + 1, carry, ys_ta
+    ys_ta = tf.nest.map_structure(
+        lambda y_ta, y: (y_ta.write(i_, y) if y is not None else y_ta),
+        ys_ta, ys)
+    i = i + 1
+    safe_carry = to_safe(carry)
+    return i, safe_carry, ys_ta
   xs_spec = tf.nest.map_structure(
-      lambda t: tf.TensorSpec(t.shape[1:], t.dtype), xs)
+      lambda t: tf.TensorSpec(t.shape[1:], t.dtype) if t is not None else None,
+      xs)
   _, ys_spec = eval_on_shapes(f)(init, xs_spec)
+  # ys_ta can't contain None because tf.while_loop doesn't allow None in
+  # loop_vars.
   ys_ta = tf.nest.map_structure(
-      lambda y: tf.TensorArray(y.dtype, size=0, dynamic_size=True), ys_spec)
-  _, carry, ys_ta = tf.while_loop(
-      lambda i, *_: i < length, body, (0, init, ys_ta))
-  ys = tf.nest.map_structure(lambda y_ta: y_ta.stack(), ys_ta)
+      lambda y: tf.TensorArray(y.dtype if y is not None else tf.float32, size=0,  # pylint: disable=g-long-lambda
+                               dynamic_size=True),
+      ys_spec)
+  safe_init = to_safe(init)
+  _, safe_carry, ys_ta = tf.while_loop(
+      lambda i, *_: i < length, body, (0, safe_init, ys_ta))
+  carry = from_safe(safe_carry)
+  def _stack(a, spec):
+    if spec is None:
+      return None
+    a = a.stack()
+    a.set_shape((length,) + a.shape[1:])
+    return a
+  ys = tf.nest.map_structure(_stack, ys_ta, ys_spec)
   return _tf_to_np((carry, ys))
 
 
@@ -949,7 +1404,7 @@ class ShardedNdArray(object):
     self.n_devices = len(tensors)
 
   def __getitem__(self, i):
-    return self.tensors[i]
+    return tf_np.asarray(self.tensors[i])
 
   @property
   def shape(self):
@@ -1009,7 +1464,7 @@ def pmap_config(axis_name, devices):
     _pmap_config.set_devices(old_devices)
 
 
-def psum(tensor, axis_name=None):
+def _psum(tensor, axis_name=None):
   """Sum all-reduction.
 
   Args:
@@ -1026,12 +1481,27 @@ def psum(tensor, axis_name=None):
   devices = _pmap_config.devices()
   if devices is None:
     raise ValueError("Can't retrieve the device list from the surrounding pmap")
+  tensor = tf_np.asarray(tensor)
   if tpu_devices(devices):
+    # TODO(b/170895907): Remove this workaround when tpu.cross_replica_sum
+    #   supports int64/float64.
+    is_int64 = False
+    is_float64 = False
+    if tensor.dtype == np.int64:
+      is_int64 = True
+      tensor = tensor.astype(np.int32)
+    elif tensor.dtype == np.float64:
+      is_float64 = True
+      tensor = tensor.astype(np.float32)
     # TODO(wangpeng): Supply the `group_assignment` argument to
-    # tpu.cross_replica_sum, calculated from `devices`.
-    return tf.compat.v1.tpu.cross_replica_sum(tensor)
+    #   tpu.cross_replica_sum, calculated from `devices`.
+    tensor = tf.compat.v1.tpu.cross_replica_sum(tensor)
+    if is_int64:
+      tensor = tf.cast(tensor, tf.int64)
+    elif is_float64:
+      tensor = tf.cast(tensor, tf.float64)
   else:
-    return tf.raw_ops.CollectiveReduce(
+    tensor = tf.raw_ops.CollectiveReduce(
         input=tensor.data,
         group_size=len(devices),
         group_key=_GROUP_KEY,
@@ -1039,6 +1509,12 @@ def psum(tensor, axis_name=None):
         merge_op="Add",
         final_op="Id",
         subdiv_offsets=(0,))
+  return tf_np.asarray(tensor)
+
+
+def psum(tensors, axis_name=None):
+  return tf.nest.map_structure(
+      functools.partial(_psum, axis_name=axis_name), tensors)
 
 
 # Note this is not available in the jax api, but seemed like a reasonable API
@@ -1087,7 +1563,12 @@ def _get_pmap_impl(f, devices, has_tpu):
   """
   if has_tpu:
     # Workaround b/121383831
-    f = _record_result_type(f)
+    output_is_list = [False]  # Use list for mutability
+    def recorder(args, kwargs, res):
+      del args, kwargs
+      output_is_list[0] = isinstance(res, list)
+      return res
+    f = _record_result_type(recorder, f)
 
   def tf_f(*tf_args):
     """A wrapper for `f` that takes/returns tensors."""
@@ -1101,7 +1582,12 @@ def _get_pmap_impl(f, devices, has_tpu):
     def fn(inputs):
       # TODO(wangpeng): Supply the `device_assignment` argument to
       # tpu.replicate, calculated from `devices`.
-      return tf.compat.v1.tpu.replicate(tf_f, inputs)
+      res = tf.compat.v1.tpu.replicate(tf_f, inputs)
+      # Workaround b/121383831
+      if (res and isinstance(res[0], list) and len(res[0]) == 1 and
+          not output_is_list[0]):
+        res = [x[0] for x in res]
+      return res
 
     return fn
   else:
@@ -1215,14 +1701,7 @@ def pmap(f, axis_name=None, devices=None):
         tensors.append(flattened_results[j][i])
       final_tree.append(ShardedNdArray(tensors))
 
-    final_actual_result = tf.nest.pack_sequence_as(results[0], final_tree)
-
-    # Workaround b/121383831
-    if (has_tpu and isinstance(final_actual_result, list) and
-        len(final_actual_result) == 1) and not _orig_result_is_list.val:
-      return final_actual_result[0]
-    else:
-      return final_actual_result
+    return tf.nest.pack_sequence_as(results[0], final_tree)
 
   return wrapper
 
@@ -1265,17 +1744,113 @@ def accelerators(devices=None):
   return tpu_devices(devices) or gpu_devices(devices)
 
 
-# TODO(agarwal): support axes arguments.
-def vmap(f):
+def _tree_broadcast(to, s):
+  """Broadcasts `s` to the nested structure `to`."""
+  if not isinstance(to, (list, tuple, dict)):
+    if not isinstance(s, (int, type(None))):
+      raise ValueError
+    return s
+  if isinstance(s, (int, type(None))):
+    return tf.nest.map_structure(lambda x: s, to)
+  if isinstance(to, (list, tuple)):
+    if len(to) != len(s):
+      raise ValueError
+    new_s = [_tree_broadcast(x, y) for x, y in zip(to, s)]
+    if isinstance(to, tuple):
+      new_s = tuple(new_s)
+    return new_s
+  elif isinstance(to, dict):
+    return {k: _tree_broadcast(to[k], s[k]) for k in to}
+  else:
+    raise TypeError("Unsupported type %s" % type(to))
+
+
+def vmap(f, in_axes=0, out_axes=0):
   """Returns a function that maps `f` over first dimension of inputs."""
+  in_axes_flat = tf.nest.flatten(in_axes)
+  if not all(isinstance(l, (type(None), int))
+             for l in in_axes_flat):
+    raise TypeError(
+        "vmap in_axes must be an int, None, or (nested) container with "
+        "those types as leaves, but got {}.".format(in_axes))
+  if all(isinstance(l, type(None)) for l in in_axes_flat):
+    raise ValueError("vmap must have at least one non-None value in in_axes")
+
+  out_axes_flat = tf.nest.flatten(out_axes)
+  if not all(isinstance(l, (type(None), int))
+             for l in out_axes_flat):
+    raise TypeError(
+        "vmap out_axes must be an int, None, or (nested) container with "
+        "those types as leaves, but got {}.".format(out_axes))
 
   def _f(*args):
-    tf_args = tf.nest.map_structure(lambda x: tf_np.asarray(x).data, args)
+    flat_args = tf.nest.flatten(args)
+    try:
+      f_in_axes = _tree_broadcast(args, in_axes)
+    except ValueError:
+      six.reraise(
+          ValueError,
+          ValueError(
+              "vmap in_axes specification must be a tree prefix of the "
+              r"corresponding value, got specification %s for value tree %s" % (
+                  in_axes, args)),
+          sys.exc_info()[2])
+    f_in_axes_flat = tf.nest.flatten(f_in_axes)
 
-    def tf_f(x):
-      return f(*x)
+    def tf_f(tf_args):
+      """Function passed to tf.vectorized_map call."""
+      # Note that unbatched arguments are not passed to tf_f. Here we fill thos
+      # arguments back before calling `f`.
+      tf_flat_args = []
+      j = 0
+      for arg, axis in zip(flat_args, f_in_axes_flat):
+        if axis is None:
+          tf_flat_args.append(arg)
+        else:
+          tf_flat_args.append(tf_args[j])
+          j += 1
+      unbatched_args = tf.nest.pack_sequence_as(args, tf_flat_args)
+      return f(*unbatched_args)
 
+    # Constructs arguments to pass to `tf_f`.
+    # Unbatch arguments are skipped. Arguments with non-zero axis are
+    # transposed.
+    tf_args = []
+    for arg, axis in zip(flat_args, f_in_axes_flat):
+      if axis is None:
+        continue
+      arg = tf_np.asarray(arg)
+      if axis != 0:
+        arg = tf_np.moveaxis(arg, axis, 0)
+      tf_args.append(arg)
+    # TODO(agarwal): consider creating a tf.function outside of _f and reusing
+    # that to avoid overheads of re-vectorizing the code when running eagerly.
     outputs = tf.vectorized_map(tf_f, tf_args)
-    return tf.nest.map_structure(tf_np.asarray, outputs)
+    try:
+      f_out_axes = _tree_broadcast(outputs, out_axes)
+    except ValueError:
+      six.reraise(
+          ValueError,
+          ValueError(
+              "vmap out_axes specification must be a tree prefix of the "
+              r"corresponding value, got specification %s for value tree %s" % (
+                  out_axes, outputs)),
+          sys.exc_info()[2])
+
+    def map_output(x, axis):
+      """Maps output of tf.vectorized_map to the final output."""
+      x = tf_np.asarray(x)
+      if axis is None:
+        # Note that `tf.vectorized_map always batches the outputs.
+        # Here we unbatch it again.
+        return x[0, ...]
+      elif axis == 0:
+        return x
+      else:
+        # Need to transpose the output.
+        return tf_np.moveaxis(x, 0, axis)
+    new_outputs = [map_output(output, axis) for output, axis in zip(
+        tf.nest.flatten(outputs), tf.nest.flatten(f_out_axes))]
+    return tf.nest.pack_sequence_as(outputs, new_outputs)
 
   return _f
