@@ -20,15 +20,19 @@ from __future__ import division
 from __future__ import print_function
 
 import functools
+import itertools
+
 from absl import flags
 from absl.testing import parameterized
 
+from jax import lax
 import numpy as np
 import tensorflow.compat.v2 as tf
 
 from trax.tf_numpy import extensions
 import trax.tf_numpy.numpy as tf_np
 
+
 FLAGS = flags.FLAGS
 
 flags.DEFINE_bool("requires_tpu", False, "Requires TPU.")
@@ -91,10 +95,14 @@ def scan_reference(f, init, xs):
   return carry, ys
 
 
+def spec(*args):
+  return tf.TensorSpec(args, tf.float32)
+
+
 class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
 
   def __init__(self, methodName="runTest"):  # pylint: disable=invalid-name
-    super(ExtensionsTest, self).__init__(methodName)
+    super().__init__(methodName)
     physical_devices = tf.config.experimental.list_physical_devices("CPU")
     tf.config.experimental.set_virtual_device_configuration(
         physical_devices[0], [
@@ -300,45 +308,81 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
     check_trace_only_once(
         tf.convert_to_tensor(value=1), tf.convert_to_tensor(value=2))
 
-  def _testEvalOnShapes(self, transformer):
+  def _testEvalOnShapes(self, transformer, allow_static_outputs):
 
-    def f(a, b):
-      return tf_np.sum(tf_np.sqrt(tf_np.exp(a)) + b)
+    # A class that's not convertable to tensor
+    class Thing:
 
-    f_prime = transformer(f)
+      def __init__(self, value):
+        self.value = value
+
+    def f(a, b, reverse=False):
+      res = tf_np.sum(tf_np.sqrt(tf_np.exp(a)) + b)
+      res = (res, 10)
+      if allow_static_outputs:
+        res = res + (Thing(20),)
+      if reverse:
+        res = tuple(reversed(res))
+      return res
+
+    f_prime = transformer(
+        f, static_argnums=(2,), allow_static_outputs=allow_static_outputs)
     shape = [10]
     dtype = np.float16
     a = tf_np.zeros(shape=shape, dtype=dtype)
     b = tf_np.zeros(shape=shape, dtype=dtype)
-    expected = f(a, b)
+    expected, *_ = f(a, b)
     got = f_prime(a, b)
-    self.assertAllEqual(expected.shape, got.shape)
-    self.assertAllEqual(expected.dtype, got.dtype)
+    def check(got):
+      self.assertIsInstance(got[0], (tf.TensorSpec, tf_np.ndarray))
+      self.assertAllEqual(expected.shape, got[0].shape)
+      self.assertAllEqual(expected.dtype, got[0].dtype)
+      if allow_static_outputs:
+        self.assertIsInstance(got[1], int)
+        self.assertEqual(10, got[1])
+        self.assertIsInstance(got[2], Thing)
+        self.assertEqual(20, got[2].value)
+      else:
+        self.assertIsInstance(got[1], (tf.TensorSpec, tf_np.ndarray))
+        self.assertAllEqual((), got[1].shape)
+    check(got)
     # Call again since the code path is different on second call
     got = f_prime(a, b)
-    self.assertAllEqual(expected.shape, got.shape)
-    self.assertAllEqual(expected.dtype, got.dtype)
+    check(got)
+    # Retrace and check again
+    got = f_prime(a, b, True)
+    check(tuple(reversed(got)))
+    got = f_prime(a, b, True)
+    check(tuple(reversed(got)))
 
-  def testEvalOnShapes(self):
+  @parameterized.named_parameters(("_%s" % b, b) for b in [False, True])
+  def testEvalOnShapes(self, allow_static_outputs):
+    self._testEvalOnShapes(extensions.eval_on_shapes, allow_static_outputs)
 
-    def transformer(f):
-      return extensions.eval_on_shapes(f)
-
-    self._testEvalOnShapes(transformer)
+  def testEvalOnShapesNested(self):
+    transformer = functools.partial(extensions.eval_on_shapes,
+                                    allow_static_outputs=True)
+    @transformer
+    def outer():
+      @transformer
+      def inner():
+        return 1
+      return inner() + 2
+    r = outer()
+    self.assertIsInstance(r, int)
+    self.assertEqual(3, r)
 
   def testJitOfEvalOnShapes(self):
     """Tests that eval_on_shapes can be called within jit."""
 
-    def transformer(f):
+    def transformer(f, **kwargs):
+      def f_prime(*args):
+        res = extensions.eval_on_shapes(f, **kwargs)(*args)
+        return tf.nest.map_structure(
+            lambda x: tf_np.zeros(x.shape, x.dtype), res)
+      return extensions.jit(f_prime, kwargs.get("static_argnums", ()))
 
-      @extensions.jit
-      def f_prime(a, b):
-        shape_dtype = extensions.eval_on_shapes(f)(a, b)
-        return tf_np.zeros(shape=shape_dtype.shape, dtype=shape_dtype.dtype)
-
-      return f_prime
-
-    self._testEvalOnShapes(transformer)
+    self._testEvalOnShapes(transformer, False)
 
   def testEvalOnShapesNoUnnecessaryTracing(self):
 
@@ -364,6 +408,106 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
     check_trace_only_once(
         tf.convert_to_tensor(value=1), tf.convert_to_tensor(value=2))
 
+  @parameterized.parameters(
+      {
+          "lhs_np": np.ones((5, 3)),
+          "rhs_np": np.ones((3, 2)),
+          "dims": (((1,), (0,)), ((), ()))
+      },
+      {
+          "lhs_np": np.ones((5, 3)),
+          "rhs_np": np.ones((5, 3)),
+          "dims": (((0, 1), (0, 1)), ((), ()))
+      },
+      {
+          "lhs_np": np.ones((5, 3, 2)),
+          "rhs_np": np.ones((2, 3, 2)),
+          "dims": (((1, 2), (1, 0)), ((), ()))
+      },
+      {
+          "lhs_np": np.ones((6, 5, 3)),
+          "rhs_np": np.ones((6, 3, 2)),
+          "dims": (((2,), (1,)), ((0,), (0,)))
+      },
+      {
+          "lhs_np": np.ones((6, 3, 5)),
+          "rhs_np": np.ones((6, 3, 2)),
+          "dims": (((1,), (1,)), ((0,), (0,)))
+      },
+      {
+          "lhs_np": np.ones((5, 3, 2, 2)),
+          "rhs_np": np.ones((5, 2, 2, 6)),
+          "dims": (((2, 3), (1, 2)), ((0,), (0,)))
+      },
+      {
+          "lhs_np": np.ones((2, 2, 5, 3)),
+          "rhs_np": np.ones((2, 2, 3, 2)),
+          "dims": (((3,), (2,)), ((0, 1), (0, 1)))
+      },
+      {
+          "lhs_np": np.ones((2, 2, 5, 2)),
+          "rhs_np": np.ones((2, 2, 3, 2)),
+          "dims": (((3,), (1,)), ((0,), (0,)))
+      },
+      {
+          "lhs_np": np.ones((2, 2, 5, 3, 3)),
+          "rhs_np": np.ones((2, 3, 2, 3, 2)),
+          "dims": (((4,), (1,)), ((0,), (0,)))
+      },
+  )
+  def test_tf_dot_general(self, lhs_np, rhs_np, dims):
+    ans = lax.dot_general(lhs_np, rhs_np, dims)
+    result = extensions.tf_dot_general(lhs_np, rhs_np, dims)
+    self.assertAllClose(result, np.array(ans))
+
+  @parameterized.named_parameters([
+      ("_lhs_shape={}_rhs_shape={}_strides={}_padding={}"  # pylint: disable=g-complex-comprehension
+       "_lhs_dilation={}_rhs_dilation={}"
+       "_feature_group_count={}_batch_group_count={}_dims={}"
+       "_perms={}".format(lhs_shape, rhs_shape,
+                          strides, padding, lhs_dilation, rhs_dilation,
+                          feature_group_count, batch_group_count, ",".join(
+                              dimension_numbers), perms),
+       lhs_shape, rhs_shape, strides, padding, lhs_dilation, rhs_dilation,
+       feature_group_count, batch_group_count, dimension_numbers, perms)
+      for batch_group_count, feature_group_count in [(1, 1)]
+      for lhs_shape, rhs_shape in [
+          ((b * batch_group_count, i * feature_group_count, 9, w),
+           (j * feature_group_count * batch_group_count, i, 4, 5))
+          for w in [0, 10]
+          for b, i, j in itertools.product([2, 3], repeat=3)]
+      for strides in [(1, 1), (2, 1)]
+      for padding in ["SAME"]
+      for lhs_dilation, rhs_dilation in [
+          (None, (1, 1))
+      ]
+      for dimension_numbers, perms in [
+          (("NHWC", "HWIO", "NHWC"), ([0, 2, 3, 1], [2, 3, 1, 0]))
+      ]])
+  def testConvGeneralDilated(self, lhs_shape, rhs_shape, strides,
+                             padding, lhs_dilation, rhs_dilation,
+                             feature_group_count, batch_group_count,
+                             dimension_numbers, perms):
+    lhs_perm, rhs_perm = perms  # permute to compatible shapes
+
+    lhs = np.transpose(np.ones(lhs_shape), lhs_perm)
+    rhs = np.transpose(np.ones(rhs_shape), rhs_perm)
+
+    jax_conv = lax.conv_general_dilated(lhs, rhs, strides, padding,
+                                        lhs_dilation, rhs_dilation,
+                                        dimension_numbers,
+                                        feature_group_count,
+                                        batch_group_count)
+
+    tf_conv = extensions.tf_conv_general_dilated(lhs, rhs, strides,
+                                                 padding, None,
+                                                 lhs_dilation, rhs_dilation,
+                                                 dimension_numbers,
+                                                 feature_group_count,
+                                                 batch_group_count)
+
+    self.assertAllEqual(tf_conv, tf_np.asarray(jax_conv))
+
   def testConv(self):
     y = extensions.conv(
         np.ones([5, 320, 480, 3], dtype=np.float32),
@@ -444,13 +588,17 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
     rng = np.random.RandomState(0)
 
     d = rng.randn(2)
-    def f(c_g, a_e):
+    def f(c_g_i, a_e_h):
+      c_g, i = c_g_i
       c, g = c_g
-      a, e = a_e
+      a, e_h = a_e_h
+      e, h = e_h
       assert a.shape == (3,)
       assert e.shape == ()  # pylint: disable=g-explicit-bool-comparison
       assert c.shape == (4,)
       assert g.shape == (2,)
+      assert i is None
+      assert h is None
       b = tf_np.cos(tf_np.sum(tf_np.sin(a)) + tf_np.sum(tf_np.cos(c)) +
                     tf_np.sum(tf_np.tan(d)))
       f = tf_np.cos(a)
@@ -458,20 +606,26 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
       g = tf_np.sin(g * b)
       assert b.shape == ()  # pylint: disable=g-explicit-bool-comparison
       assert f.shape == (3,)
-      return [c, g], (b, f)
+      return [(c, g), i], (b, [f, h])
 
-    xs = (rng.randn(5, 3), rng.randn(5))
-    init = [rng.randn(4), rng.randn(2)]
+    xs = (rng.randn(5, 3), [rng.randn(5), None])
+    init = [(rng.randn(4), rng.randn(2)), None]
 
-    c_g, b_f = extensions.scan(f, init, xs)
-    self.assertIsInstance(c_g, list)
-    self.assertIsInstance(b_f, tuple)
+    c_g_i, b_f_h = extensions.scan(f, init, xs)
+    self.assertIsInstance(c_g_i, list)
+    self.assertIsInstance(b_f_h, tuple)
+    c_g, i = c_g_i
     c, g = c_g
-    b, f = b_f
+    self.assertIsInstance(c_g, tuple)
     self.assertEqual((4,), c.shape)
     self.assertEqual((2,), g.shape)
+    self.assertIsNone(i)
+    b, f_h = b_f_h
+    f, h = f_h
+    self.assertIsInstance(f_h, list)
     self.assertEqual((5,), b.shape)
     self.assertEqual((5, 3), f.shape)
+    self.assertIsNone(h)
 
   @parameterized.named_parameters(
       (f"_{jit_scan}_{jit_f}", jit_scan, jit_f)  # pylint: disable=g-complex-comprehension
@@ -517,6 +671,19 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
         to_tf_fn(functools.partial(losses, scan)), (c, xs))
     self.assertAllClose(theoretical, numerical, atol=1e-3, rtol=3e-4)
 
+  @parameterized.named_parameters(
+      (f"_{i}", *args)  # pylint: disable=g-complex-comprehension
+      for i, args in enumerate([
+          (lambda c, x: (c + 1, tf_np.sum(c + x, 0)),
+           [spec(2), spec(4, 3, 2)], [spec(2), spec(4, 2)]),
+          (lambda c, x: (c + 1, tf_np.sum(c + x, 0)),
+           [spec(2), spec(0, 3, 2), 0], [spec(2), spec(0, 2)]),
+      ]))
+  def testScanShape(self, f, inputs, expected_outputs):
+    outputs = extensions.eval_on_shapes(
+        functools.partial(extensions.scan, f), static_argnums=(2,))(*inputs)
+    self.assertAllEqual(expected_outputs, outputs)
+
   def testPrng(self):
     self.assertAllEqual(tf_np.asarray(123, np.int64), extensions.prng(123))
 
@@ -694,6 +861,28 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
     self.assertAllClose(result[0], 4)
     self.assertAllClose(result[1], 4)
 
+  def testPsumStruct(self):
+    devices = self._get_two_devices(require_same_type=True)
+
+    def reduce_sum(a):
+      a = extensions.psum(a)
+      tf.nest.map_structure(
+          lambda x: self.assertIsInstance(x, tf_np.ndarray), a)
+      return a
+
+    data = [tf_np.asarray([1, 3]), tf_np.asarray([2, 4], np.int64)]
+    pmapped = extensions.pmap(reduce_sum, devices=devices)
+    result = pmapped(data)
+
+    self.assertIsInstance(result[0][0], tf_np.ndarray)
+    self.assertIsInstance(result[0][1], tf_np.ndarray)
+    self.assertIsInstance(result[1][0], tf_np.ndarray)
+    self.assertIsInstance(result[1][1], tf_np.ndarray)
+    self.assertAllClose(result[0][0], 4)
+    self.assertAllClose(result[0][1], 4)
+    self.assertAllClose(result[1][0], 6)
+    self.assertAllClose(result[1][1], 6)
+
   def testPmean(self):
     if extensions.tpu_devices():
       self.skipTest("pmean for TPU is not supported yet")
